import os
import urllib
import urllib.request


def download(url, filename):
    """
    Retrieve a file from the web and save it to a location

    Parameters
    ----------
    url : String
        The url of the file to download

    filename :
        The location of where to save it

    Returns
    -------
    void
    """

    urllib.request.urlretrieve(url, filename)


def wget_fromurl(wgeturl, targetdir, verbose=True):
    """
    Run a wget script to download files to a target dir

    Parameters
    ----------
    wgeturl : string
        url of the wget file

    targetdir : string
        Target directory for saving the files

    verbose : bool
        print status updates?

    Returns
    -------
    void
    """

    # to use locally replace: "with open(wgeturl)"
    with urllib.request.urlopen(wgeturl) as fl:
        for rowb in fl:  # rows come in as binary
            row = rowb.decode('utf-8')  # comes back as a bytestream
            # just get the URL to the file
            u = row.split()[-1].strip()
            # just get the filename portion
            fn = os.path.split(u)[-1]
            # Download the file
            if not os.path.isfile(os.path.join(targetdir, fn)):
                if verbose:
                    print(fn)
                download(u, os.path.join(targetdir, fn))
            else:
                if verbose:
                    print(fn + " exists, skipping...")


def ensurepath(path):
    """
    Test if a path exists, if not, create it
    Parameters
    ----------
    path : str
        The path to test

    Returns
    -------
        Void
    """
    if not os.path.exists(path):
        os.makedirs(path)


def getfiles(path, ext=None, sort=True):
    """
    Get a list of all files on a path, optionally filtering by ext and sorting
    Parameters
    ----------
    path : str
        the path to search
    ext : str
        the extension to match as string (e.g. ".txt").
        If None, return all files
    sort : bool
        should the files be sorted?

    Returns
    -------
    files : list
        A list of the matching files found in the directory
    """
    baselist = os.listdir(path)
    files = []

    # Split extensions if desired
    if ext is not None:
        for fn in baselist:
            if os.path.splitext(fn)[-1] == ext:
                files.append(fn)
    else:
        files = baselist

    # Sort if desired
    if sort:
        files.sort()

    return files



import numpy as np
from scipy.interpolate import RegularGridInterpolator

import pvlib

import matplotlib.pyplot as plt
from scipy.ndimage import sobel, uniform_filter

from solarspatialtools.stats import variability_score


def _random_at_scale(rand_size, final_size, plot=False):
    """
    Generate a 2D array of random at an initial size and linearly interpolate up to a larger final size.

    Parameters
    ----------
    rand_size : tuple
        The size of the random field
    final_size : tuple
        The size of the final interpolated field
    plot : bool
        Whether to plot the results

    Returns
    -------
    random : np.ndarray
        The original random field
    random_new : np.ndarray
        The interpolated random field
    """

    # Generate random values at the scale of rand_size
    random = np.random.rand(np.max([rand_size[0],1]), np.max([rand_size[1],1]))

    # If random has any singleton dimensions, repeat it, because interpolation
    # requires at least 2 points in each dimension
    if random.shape[0] == 1:
        random = np.repeat(random, 2, axis=0)
    if random.shape[1] == 1:
        random = np.repeat(random, 2, axis=1)

    # Linearly interpolate to the final size
    x = np.linspace(0, 1, random.shape[0])
    y = np.linspace(0, 1, random.shape[1])

    xnew = np.linspace(0, 1, final_size[0])
    ynew = np.linspace(0, 1, final_size[1])

    # # Latest Recommended Scipy Method
    interp_f = RegularGridInterpolator((x, y), random, method='linear')
    Xnew, Ynew = np.meshgrid(xnew, ynew, indexing='ij')
    random_new = interp_f((Xnew, Ynew))

    # # # Alternate Scipy Method
    # interp_f = RectBivariateSpline(x, y, random, kx=1, ky=1)
    # # interp_ft = lambda xnew, ynew: interp_f(xnew, ynew).T
    # random_new = interp_f(xnew, ynew)

    # # # Different potentially faster Scipy Method
    # Xnew, Ynew = np.meshgrid(xnew, ynew, indexing='ij')
    # Xnew = Xnew*len(x)
    # Ynew = Ynew*len(y)
    # random_new = map_coordinates(random, [Xnew.ravel(), Ynew.ravel()], order=1, mode='nearest').reshape(final_size)

    if plot:
        # generate side by side subplots to compare
        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        axs[0].imshow(random, extent=(0, rand_size[1], 0, rand_size[0]))
        axs[0].set_title('Original Random')
        axs[1].imshow(random_new, extent=(0, final_size[1], 0, final_size[0]))
        axs[1].set_title('Interpolated Random')
        plt.show()

    return random, random_new


def _calc_vs_weights(scales, vs):
    """
    Calculate the weights for each scale based on the variability score

    Parameters
    ----------
    scales : np.ndarray
        The scales of the wavelets
    vs : float
        The variability score

    Returns
    -------
    weights : np.ndarray
        The weights for each scale
    """

    if vs < 0:
        raise ValueError("VS must be greater than 0.")

    VS1 = -np.log(1-vs/180)/0.6
    weight = scales ** (1/VS1)
    return weight / np.sum(weight)


def _calc_wavelet_weights(waves):
    """
    Calculate the weights for each wavelet scale based on the wavelet values.

    Parameters
    ----------
    waves : np.ndarray
        The wavelet values. Should be the timeseries for each mode. Consider
        generating from `pvlib.scaling._compute_wavelet`.

    Returns
    -------
    weights : np.ndarray
        The weights for each wavelet scale
    """
    scales = np.nanmean(waves**2, axis=1)
    return scales / np.sum(scales)


def _stack_random_field(weights, scales, size, normalize=False, plot=False):
    """
    Generate a field of random clouds at different scales and weights based on Perlin Noise.
    This runs relatively slowly for large arrays.

    Parameters
    ----------
    weights : np.ndarray
        The weights for each scale. See _calc_wavelet_weights or _calc_vs_weights.
        Should add to unity.
    scales : np.ndarray
        The scales for each weight. Should be the same length as weights.
    size : tuple
        The size of the field
    normalize : bool
        Whether to normalize the field to 0-1
    plot : bool
        Whether to plot the field

    Returns
    -------
    field : np.ndarray
        The field of random clouds
    """

    if len(weights) != len(scales):
        raise ValueError("Number of weights must match scales.")

    # Calculate a field for each scale, and add them together in a weighted manner to form the field
    field = np.zeros(size, dtype=float)

    for scale, weight in zip(scales, weights):
        prop = 2.0**(-scale+1)  # proportion for this scale
        xsz = int(size[0]*prop)
        ysz = int(size[1]*prop)
        _, i_field = _random_at_scale((xsz, ysz), size)
        field += i_field * weight

    # Optionally Scale it zero to 1
    if normalize:
        field = (field - np.min(field))
        field = field / np.max(field)
        assert np.min(field) == 0
        assert np.max(field) == 1

    if plot:
        # Plot the field
        plt.imshow(field, extent=(0, size[1], 0, size[0]))
        plt.show()

    return field

def _calc_clear_mask(field, clear_frac=0.5, plot=False):
    """
    Find the value in the field that will produce an X% clear sky mask. The
    mask is 1 where clear, and 0 where cloudy.

    Parameters
    ----------
    field : np.ndarray
        The field of clouds
    clear_frac : float
        The fraction of the field that should be clear
    plot : bool
        Whether to plot the field

    Returns
    -------
    field_mask : np.ndarray
        The clear sky mask. Zero indicates cloudy, one indicates clear
    """
    if clear_frac > 1 or clear_frac < 0:
        raise ValueError("Clear fraction must be between 0 and 1.")

    # Zero where clouds, 1 where clear

    if clear_frac == 0:
        field_mask = np.zeros_like(field)

    else:
        # clipping needs to be based on pixel fraction, which thus needs to be
        # done on quantile because the field has a normal distribution
        quant = np.quantile(field, clear_frac)

        # Find that quantile and cap it
        field_mask = np.ones_like(field)
        field_mask[field > quant] = 0

    # Test to make sure that we're close to matching the desired fraction
    assert (np.isclose(clear_frac, np.sum(field_mask) / field.size, rtol=1e-3))

    if plot:
        plt.imshow(field_mask, extent=(0, field.shape[1], 0, field.shape[0]))
        plt.show()

    return field_mask

def _find_edges(base_mask, size, binarize_threshold=0, plot=False):
    """
    Find the edges of a mask using a Sobel filter and then broadens it with a uniform filter.

    Parameters
    ----------
    base_mask : np.ndarray
        The mask to find the edges of
    size : int
        The size of the uniform filter (effectively doubles the size of the filter.
    binarize_threshold : float
        The lower threshold to binarize the smoothed mask. Increase to reduce
        the overall coverage of the edges, e.g. for highly sparse areas. Scaled
        as a percentage of the smoothed mask dynamic range.

    Returns
    -------
    edges : np.ndarray
        The edges of the mask
    smoothed_binary : np.ndarray
        The smoothed binary mask
    """

    # This gets us roughly 50% overlapping with mask and 50% outside
    edges = np.abs(sobel(base_mask,0)**2+sobel(base_mask,1)**2)
    smoothed = uniform_filter(edges, size=size)

    # We want to binarize it
    smoothed[smoothed < 1e-5] = 0  # Zero out the small floating point values
    # Calculate a threshold based on quantile, because otherwise we get the whole clouds
    smoothed_binary = smoothed/np.max(smoothed) > binarize_threshold

    if plot:
        xsiz, ysiz = base_mask.shape
        # Compare the edges and uniform filtered edges side by side
        fig, axs = plt.subplots(1, 3, figsize=(10, 5))
        axs[0].imshow(base_mask, extent=(0, ysiz, 0, xsiz))
        axs[0].set_title('Original Mask')
        axs[1].imshow(edges, extent=(0, ysiz, 0, xsiz))
        axs[1].set_title('Edges')
        axs[2].imshow(smoothed_binary, extent=(0, ysiz, 0, xsiz))
        axs[2].set_title('Uniform Filtered Edges')
        plt.show()

    return edges, smoothed_binary


def _scale_field_lave(field, clear_mask, edge_mask, ktmean, ktmax=1.4, kt1pct=0.2, max_quant=0.99, plot=False):
    """
    Scale a field of clouds to match a desired mean and distribution of kt values. Follows the method used by the
    original author.

    Parameters
    ----------
    field : np.ndarray
        The field of clouds
    clear_mask : np.ndarray
        The clear sky mask. Zero indicates cloudy, one indicates clear
    edge_mask : np.ndarray
        The edge mask. Zero indicates not an edge, one indicates an edge
    ktmean : float
        The desired mean of the kt values
    ktmax : float
        The maximum of the kt values
    kt1pct : float
        The 1st percentile of the kt values
    max_quant : float
        The quantile to use for the maximum of the field
    plot : bool
        Whether to plot the results

    Returns
    -------
    clouds5 : np.ndarray
        The scaled field of clouds
    """

    # ##### Shift values of kt to range from 0.2 - 1

    # Calc the "max" and "min", excluding clear values
    field_max = np.quantile(field[clear_mask == 0], max_quant)

    # Create a flipped version of the distribution that scales between
    # slightly below kt1pct and bascially (1-field_min). I think the intent
    # here would be to make it vary between kt1pct and 1, but that's not
    # quite what it does.
    clouds3 = 1 - field*(1-kt1pct)/field_max

    # # Clip limits to sensible boundaries
    clouds3[clouds3 > 1] = 1
    clouds3[clouds3 < 0] = 0

    # ##### Calculate the cloud edge distribution #####
    ce = 1 + (clouds3 - np.min(clouds3)) / (np.max(clouds3) - np.min(clouds3)) * (ktmax - 1)

    # Calculate a scaling factor on the cloudy region, to make it match the
    # specified ktmean from the timeseries
    try:
        # Where is it neither clear nor edge
        cloud_mask = np.bitwise_or(clear_mask > 0, edge_mask) == 0
    except TypeError:
        # Where is it neither clear nor edge
        cloud_mask = np.bitwise_or(clear_mask > 0, edge_mask > 0) == 0

    # Mean scaled over whole field
    tgtsum = field.size * ktmean

    # Shifting target to exclude fully clear vals and the cloud enhancement
    diff_sum = (tgtsum
                - np.sum(clear_mask)
                - np.sum(ce[np.bitwise_and(edge_mask > 0, clear_mask==0)]))

    # Find average required in areas where it's neither cloud nor edge
    tgt_cloud_mean = diff_sum / np.sum(cloud_mask)

    # Actual current cloud mean in the field
    current_cloud_mean = np.mean(clouds3[cloud_mask])

    # Only do this scaling if we have a positive difference
    if diff_sum > 0:
        clouds4 = tgt_cloud_mean / current_cloud_mean * clouds3
    else:
        clouds4 = clouds3.copy()

    clouds5 = clouds4.copy()

    # Mask in edges and clear. Order matters. By doing clear second,
    # we ensure that the clear sky is respected over the edge mask
    clouds5[edge_mask > 0] = ce[edge_mask > 0]
    clouds5[clear_mask > 0] = 1

    if plot:
        # plt.hist(ce.flatten(), bins=100)
        # plt.hist(clouds3.flatten(), bins=100, alpha=0.5)
        # plt.hist(clouds4.flatten(), bins=100, alpha=0.5)
        # plt.hist(clouds5.flatten(), bins=100, alpha=0.5)
        # plt.hist(field.flatten(), bins=100, alpha=0.5)
        # plt.legend(["Cloud Enhancement", "1st Scaled Cloud Distribution", "2nd Scaled Cloud Distribution", "Fully Remapped Distribution",
        #             "Original Field Distribution"])
        plt.hist(clouds5.flatten()[clouds5.flatten()<1], bins=100, alpha=0.5)
        plt.hist(clouds5.flatten()[clouds5.flatten()==1], bins=100, alpha=0.5)
        plt.hist(clouds5.flatten()[clouds5.flatten()>1], bins=100, alpha=0.5)
        plt.legend(["Cloudy Area", "Clear Sky", "Edge Enhancement"])
        plt.ylabel('Frequency')
        plt.xlabel('kt')

        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        axs[0].imshow(field, extent=(0, field.shape[1], 0, field.shape[0]))
        axs[0].set_title('Original Field')
        axs[1].imshow(clouds5, extent=(0, field.shape[1], 0, field.shape[0]))
        axs[1].set_title('Scaled Field w/ Cloud Enhancement')
        plt.show()

    return clouds5

def _scale_field_basic(field, clear_mask, edge_mask, ktmean, ktmax=1.4, kt1pct=0.2, quant=0.999, flipdistr=False, cs_smoothing=None, plot=False):
    """
    Scale a field of clouds to match a desired mean and distribution of kt values. Uses an alternate method that tries
    to more reliably match the cloud enhancement and clear sky values.

    Parameters
    ----------
    field : np.ndarray
        The field of clouds
    clear_mask : np.ndarray
        The clear sky mask. Zero indicates cloudy, one indicates clear
    edge_mask : np.ndarray
        The edge mask. Zero indicates not an edge, one indicates an edge
    ktmean : float
        The desired mean of the kt values
    ktmax : float
        The maximum of the kt values
    kt1pct : float
        The 1st percentile of the kt values
    quant : float
        The quantile to use as a surrogate for the max and min limits of the field.
    flipdistr : bool
        Whether to flip the PDF distribution of the clearsky that serves as the basis
    cs_smoothing : float
        The minimum value of the distribution desired for the clearsky values. If None, no smoothing is applied and
        clearsky locations will have a uniform 1 value of CSI.
    plot : bool
        Whether to plot the results

    Returns
    -------
    clouds5 : np.ndarray
        The scaled field of clouds"""

    # flipdistr = True
    # cs_smoothing = 0.6

    field_max = np.quantile(field[clear_mask == 0], quant)
    field_min = np.quantile(field[clear_mask == 0], 1-quant)

    # cloud field with max and min percentiles running from 0 - 1
    clouds_fr = (field - field_min) / (field_max - field_min)

    # Rather than clipping, we'll reflect the portions outside the range
    clouds_fr[clouds_fr > 1] = 2 - clouds_fr[clouds_fr > 1]
    clouds_fr[clouds_fr < 0] = -clouds_fr[clouds_fr < 0]

    if flipdistr:
        # flip the distribution so that it runs from 1 to 0
        clouds_fr = 1 - clouds_fr

    # create a cloud enhancement field that scales from 1 to ktmax
    ce = clouds_fr * (ktmax - 1) + 1

    # calculate the cloud mask
    cloud_mask = np.bitwise_or(clear_mask > 0, edge_mask > 0) == 0

    # Create a placeholder
    clouds5 = np.zeros_like(field)
    # copy in the edges
    clouds5[edge_mask > 0] = ce[edge_mask > 0]
    # copy in the clear sky
    if cs_smoothing is not None:
        clouds5[clear_mask > 0] = clouds_fr[clear_mask > 0] * (1 - cs_smoothing) + cs_smoothing
    else:
        clouds5[clear_mask > 0] = 1

    # calculate the required sum of the cloudy region
    tgtsum = field.size * ktmean
    tgt_cloud_sum = tgtsum - np.sum(clouds5)

    # Subtract off the baseline of the minimum kt
    baseline_sum = kt1pct * np.sum(cloud_mask)
    tgt_cloud_sum = tgt_cloud_sum - baseline_sum

    # Calculate the current sum of the cloudy region and scale if
    # appropriate. We don't want to scale up because the clouds would
    # exceed kt = 1.
    current_cloud_sum = np.sum(clouds_fr[cloud_mask])

    if 0 < tgt_cloud_sum < current_cloud_sum and current_cloud_sum > baseline_sum:
        clouds5[cloud_mask] = clouds_fr[cloud_mask] * tgt_cloud_sum / current_cloud_sum + kt1pct
    else:
        print("Warning! Can't match desired statistics.")
        # convert clouds_fr to run from kt1pct to 1 and assign values
        clouds5[cloud_mask] = clouds_fr[cloud_mask] * (1 - kt1pct) + kt1pct

    if plot:
        plt.hist(clouds5.flatten()[clouds5.flatten()<1], bins=100, alpha=0.5)
        plt.hist(clouds5.flatten()[clouds5.flatten()==1], bins=100, alpha=0.5)
        plt.hist(clouds5.flatten()[clouds5.flatten()>1], bins=100, alpha=0.5)
        plt.legend(["Cloudy Area", "Clear Sky", "Edge Enhancement"])
        plt.ylabel('Frequency')
        plt.xlabel('kt')

        fig, axs = plt.subplots(1, 2, figsize=(10, 5))
        axs[0].imshow(field, extent=(0, field.shape[1], 0, field.shape[0]))
        axs[0].set_title('Original Field')
        axs[1].imshow(clouds5, extent=(0, field.shape[1], 0, field.shape[0]))
        axs[1].set_title('Scaled Field w/ Cloud Enhancement')
        plt.show()

    return clouds5

def get_timeseries_stats(kt_ts, clear_threshold=0.95, plot=False):
    """
    Get the properties of a time series of kt values.

    Parameters
    ----------
    kt_ts : pandas.Series
        The time series of kt values
    clear_threshold : float
        The threshold in kt for what is considered clear
    plot : bool
        Whether to plot the wavelets

    Returns
    -------
    ktmean : float
        The mean of the time series
    kt_1_pct : float
        The 1st percentile of the time series
    ktmax : float
        The maximum of the time series
    frac_clear : float
        The fraction of clear sky
    vs : float
        The variability score
    weights : np.ndarray
        The wavelet weights
    scales : list
        The timescales of the wavelets
    """

    # Get the mean and standard deviation of the time series
    ktmean = np.mean(kt_ts)  # represents mean of kt
    ktstd = np.std(kt_ts)
    ktmax = np.max(kt_ts)  # represents peak cloud enhancement
    ktmin = np.min(kt_ts)

    kt_1_pct = np.nanquantile(kt_ts, 0.01)  # represents "lowest" kt

    # Get the fraction of clear sky with a threshold
    frac_clear = np.sum(kt_ts > clear_threshold) / kt_ts.size

    vs = variability_score(kt_ts) * 1e4

    # Compute the wavelet weights
    # should be the mean(wavelet squared) for all modes except the steady mode
    waves, tmscales = pvlib.scaling._compute_wavelet(kt_ts)

    if plot:
        # create a plot where each of the timeseries in waves is aligned vertically in an individual subplot
        fig, axs = plt.subplots(len(waves), 1, figsize=(10, 2 * len(waves)), sharex=True)
        for i, wave in enumerate(waves):
            axs[i].plot(wave)
            axs[i].set_title(f'Wavelet {i+1}')
        plt.show()

    waves = waves[:-1, :]  # remove the steady mode
    scales = [i+1 for i, _ in enumerate(tmscales[:-1])]
    weights = _calc_wavelet_weights(waves)

    return ktmean, kt_1_pct, ktmax, frac_clear, vs, weights, scales


def get_positional_ts(tgt_position, field, cloud_speed, duration=3600, pixres=1):
    """
    Get a time series of kt values for a target position in the field.

    Parameters
    ----------
    tgt_position : tuple
        The spatial target position in the field (in meters)
    field : np.ndarray
        The field of clouds generated by cloudfield_timeseries
    cloud_speed : float
        The speed of the clouds in meters per second
    duration : int
        The duration of the time series in seconds
    pixres : int
        The resolution of the field in seconds

    Returns
    -------
    sim_kt : np.ndarray
        The simulated time series of kt values for the point within the field
    """

    # Convert to the temporal coordinate system
    x = int(tgt_position[0]/cloud_speed)
    y = int(tgt_position[1]/cloud_speed)
    sim_kt = field[x:x + int(duration) + 1, y]

    return sim_kt



def cloudfield_timeseries(weights, scales, size, frac_clear, ktmean, ktmax, kt1pct, scaling='original', edgesmoothing=3):
    """
    Generate a time series of cloud fields based on the properties of a time series of kt values. This is an
    implementation of the method described by Lave et al [1]. Some aspects of the implementation diverge slightly from
    the initial paper to follow a subsequent code implementation of the method shared by the original authors.

    [1] Matthew Lave, Matthew J. Reno, Robert J. Broderick, "Creation and Value of Synthetic High-Frequency Solar Inputs
    for Distribution System QSTS Simulations," 2017 IEEE 44th Photovoltaic Specialist Conference (PVSC),
    Washington, DC, USA, 2017, pp. 3031-3033, doi: https://dx.doi.org/10.1109/PVSC.2017.8366378.

    Parameters
    ----------
    weights : np.ndarray
        The wavelet weights at each scale
    scales : list
        The scales of the wavelets, should be integer values interpreted as 2**(scale-1) seconds
    size : tuple
        The size of the field to generate, x by y
    frac_clear : float
        The fraction of clear sky
    ktmean : float
        The mean of the kt values
    ktmax : float
        The maximum of the kt values
    kt1pct : float
        The 1st percentile of the kt values
    scaling : str
        The scaling method to use. Either 'original' or 'basic'
    edgesmoothing : int
        The size of the uniform filter for edge smoothing

    Returns
    -------
    field_final : np.ndarray
        The final field of simulated clouds
    """
    cfield = _stack_random_field(weights, scales, size, plot=False)
    clear_mask = _stack_random_field(weights, scales, size)
    clear_mask = _calc_clear_mask(clear_mask, frac_clear, plot=False)  # 0 is cloudy, 1 is clear

    edges, smoothed = _find_edges(clear_mask, edgesmoothing)

    if scaling == 'original':
        field_final = _scale_field_lave(cfield, clear_mask, edges, ktmean, ktmax, kt1pct, plot=False)
    elif scaling == 'basic':
        field_final = _scale_field_basic(cfield, clear_mask, smoothed, ktmean, ktmax, kt1pct, plot=False)
    else:
        raise ValueError("Scaling method must be either 'original' or 'basic'.")

    return field_final



import pvlib
import numpy as np
import pandas as pd


def clearsky_index(ghi, clearsky_ghi, max_clearsky_index=2.0):
    """
    Wrapper for pvlib.irradiance.clearsky_index. Original doesn't handle some
    of the datatypes that we use here cleanly.

    The clearsky index is the ratio of global to clearsky global irradiance.
    Negative and non-finite clearsky index values will be truncated to zero.

    Parameters
    ----------
    ghi : numeric, pandas.Series, pandas.DataFrame
        Global horizontal irradiance in W/m^2.

    clearsky_ghi : numeric
        Modeled clearsky GHI

    max_clearsky_index : numeric, default 2.0
        Maximum value of the clearsky index. The default, 2.0, allows
        for over-irradiance events typically seen in sub-hourly data.

    Returns
    -------
    clearsky_index : numeric
        Clearsky index
    """

    if isinstance(ghi, pd.DataFrame):  # apply column by column
        return ghi.apply(lambda x:
                         pvlib.irradiance.clearsky_index(x, clearsky_ghi, 2),
                         axis=0)
    else:
        return pvlib.irradiance.clearsky_index(ghi, clearsky_ghi,
                                           max_clearsky_index)



import numpy as np
from scipy.optimize import minimize_scalar, leastsq

from solarspatialtools.signalproc import compute_delays
from solarspatialtools.spatial import project_vectors, compute_vectors
from solarspatialtools.spatial import pol2rect, magnitude, dot

from scipy.optimize import linear_sum_assignment, shgo
from scipy.stats import linregress

from enum import Enum
import itertools


class Flag(Enum):
    GOOD = 0
    NOSIGNAL = 1
    LOWVAR_S = 2
    LOWCCF = 3
    OVERCAST = 4
    NOLAG = 5
    LOWCORR = 6
    STATS = 7
    VELLIMIT = 8
    ALLBAD = 9
    TREND = 10
    SAME = 11

    DESCS = {
        GOOD: "good data",
        NOSIGNAL: "one signal has zero variance, likely bogus",
        LOWVAR_S: "Low variation ratio in a signal",
        LOWCCF: "Low cloud cover fraction in a signal (i.e. too clear)",
        OVERCAST: "low overall kt in both signals",
        NOLAG: "no lag computed (infinite velocity)",
        LOWCORR: "low lagging correlation (<0.8)",
        VELLIMIT: "Jamaly and Kleissl Low velocity limit",

        ALLBAD: "no good datapoints",
        TREND: "trend",
        SAME: "pair is same site"}


class WindspeedData:
    """
    A holder data structure
    """
    corr_raw = None  # Zero-lag cross-correlations relative to ref point
    corr_lag = None  # Max-lag cross-correlation
    allpairs = None  # Points used for windspeed measurement
    pair_lag = None  # Point-wise maximum lag
    pair_dists = None  # Point-wise distance to reference in m
    wind_speed = None  # Wind speed in m/s
    wind_angle = None  # Wind dir in radians
    pair_flag = None
    flag = None  # A data quality flag
    method_data = None  # A holder for method-specific data


def _get_pairs(all_points, must_contain=None, replacement=True):
    """
    Build the array of all possible site pairs from a given list of points.
    Works by generating number of combinations (with replacement)

    Parameters
    ----------
    all_points : iterable
        List of all point ids

    must_contain : iterable (optional, default: None)
        List of points that must be contained within all entries of the final
        list. Use for example to force all combinations to contain a reference
        site.

    replacement : boolean (optional, default: True)
        Should the list of pairs be executed with replacement? True will result
        in the identity pair (i.e. 1-1) being generated, while False will lead
        to only non-identical pairs.

    Returns
    -------
    pairs : np.array
        An N x 2 numpy array containing all the combination pairs.

    """

    # # Old Method - All possible permutations (2-way)
    # pairs_all = list(itertools.permutations(all_points, 2))
    # # Permutations with replacement
    # pairs_all = list(itertools.product(all_points, repeat=2))

    # New method, combination not permutation
    if replacement:
        pairs_all = list(
            itertools.combinations_with_replacement(all_points, 2))
    else:
        pairs_all = list(itertools.combinations(all_points, 2))

    if must_contain is not None:
        # Downselect to those that contain the parts
        pairs = []
        for pair in pairs_all:
            if pair[0] in must_contain or pair[1] in must_contain:
                pairs.append(pair)
        pairs = np.array(pairs)
        return np.array(pairs)
    else:
        return np.array(pairs_all)


def compute_cmv(timeseries, positions, reference_id=None, method="jamaly",
                options=None):
    """
    Find Cloud Motion Vector based on clear sky index timeseries from a cluster
    of sensors.

    Default uses the method by Jamaly and Kleissl [1]. An alternate method
    described by Gagne [2] is available. Optionally computes relative
    to a single reference point rather than a global computation across all
    possible site pairs.

    Parameters
    ----------
    timeseries : pandas.DataFrame
        dataframe of kt timeseries for all sensors. Columns should be labelled
        with the sensor id. All sensors should share a uniform time index.

    positions : pandas.DataFrame
        dataframe of the site positions. Index must be the site IDs. The first
        column should be the x coordinate of each site, and the second column
        should be the y coordinate. Coordinates must represent a rectilinear
        coordinate system, measured in meters. Consider converting to UTM using
        spatial.latlon2utm().

    reference_id : numeric, str, iterable, or None (default None)
        The identifier of a single reference site id within the dataframe.
        OR
        a list/tuple of identifiers of references within the dataframe.
        OR
        None to signify that all possible site pairs should be considered.

    method : str (default 'jamaly')
        Method to use. Currently accepted methods are 'jamaly' and 'gagne'

    options : dict (default {})
        Dictionary of detailed QC arguments for the methods.
            All Methods:
                ktlim : float (default 0.4)
                    Minimum permissible clear sky index for time period
            Jamaly:
                minvelocity : float (default 0 m/s)
                    Minimum permissible pairwise velocity in m/s
                maxvelocity : float (default 70 m/s)
                    Maximum permissible pairwise velocity in m/s
                var_s_lim : float (default 0.05)
                    Minimum permissible pairwise variation ratio in signals
                mincorr : float (default 0.8)
                    Minimum permissible correlation coefficient for site pairs
            Gagne:
                No method specific options

    Returns
    -------
    cmv_vel : numeric
        The cloud motion vector magnitude in meters per second

    cmv_theta : numeric
        The cloud motion direction as an angle measured CCW from east in rads

    outdata : WindspeedData
        An object containing detailed data about the individual sites

    [1] M. Jamaly and J. Kleissl, “Robust cloud motion estimation by
    spatio-temporal correlation analysis of irradiance data,” Solar
    Energy, vol. 159, pp. 306–317, Jan. 2018. [Online]. Available:
    https://www.sciencedirect.com/science/article/pii/S0038092X17309556

    [2] A. Gagné, N. Ninad, J. Adeyemo, D. Turcotte, and S. Wong, “Directional
    Solar Variability Analysis,” in 2018 IEEE Electrical Power and Energy
    Conference (EPEC) (2018) pp. 1–6, iSSN: 2381-2842
    https://www.researchgate.net/publication/330877949_Directional_Solar_Variability_Analysis
    """

    # Validate method
    methods = ['jamaly', 'gagne']
    method = method.lower()
    if method not in methods:
        raise ValueError('Method must be one of: ' + str(methods) + '.')

    # Ignore some numpy printouts, we'll deal with them manually
    np.seterr(divide='ignore')
    np.seterr(invalid='ignore')

    # Find all the needed info
    nsamples, _ = timeseries.shape
    point_ids = positions.index.to_list()

    # Set ref_ids to a list of ref_ids that should be used
    if reference_id is not None:
        # Check if it's already an iterable (not raw str though)
        if (hasattr(reference_id, "__iter__") and
                (not isinstance(reference_id, str))):
            ref_ids = reference_id
        else:
            ref_ids = [reference_id]
    else:
        # If it's none, get the full list
        ref_ids = point_ids.copy()

    # Get all possible point pairs
    pairs = _get_pairs(point_ids, ref_ids, replacement=True)

    # Get some parameters about the time series
    ts = timeseries  # Shorten the name for simplicity

    # This is the alternative to looping over all the pairs
    A = ts[pairs[:, 0]]
    B = ts[pairs[:, 1]]

    # We always need to calculate the correlations using correlation coeff.
    # scaling, because it matters for various QC checks.
    delay, extras = compute_delays(A, B, 'loop', scaling='coeff')
    corr_lag = extras['peak_corr']
    corr_mean = extras['mean_corr']

    # Vectors
    vectors_cart = (positions.loc[pairs[:, 1]].values
                    - positions.loc[pairs[:, 0]].values)

    # Pairwise QC
    pair_flags, method_out = _pairwise_qc(pairs, magnitude(vectors_cart.T),
                                          [A, B], delay, corr_lag, corr_mean,
                                          method, options)

    # Perform the Gagne QC, which looks for correlation on the whole set to
    # decide which pairs are worth using
    overall_flag = Flag.GOOD
    if method == 'gagne':
        corr_inds = corr_lag >= 0.99
        if sum(corr_inds[pair_flags == Flag.GOOD]) < 12:
            corr_inds = corr_lag >= 0.95
        if sum(corr_inds[pair_flags == Flag.GOOD]) < 12:
            corr_inds = corr_lag >= 0.9
        if sum(corr_inds[pair_flags == Flag.GOOD]) < 12:
            overall_flag = Flag.ALLBAD
        pair_flags[np.bitwise_not(corr_inds)] = Flag.LOWCORR

    # Jump out if no pairs passed QC
    if np.sum(pair_flags == Flag.GOOD) <= 1:  # they all fail QC
        overall_flag = Flag.ALLBAD
        w = WindspeedData()
        w.allpairs = pairs
        w.flag = overall_flag
        w.pair_flag = pair_flags
        return 0, 0, w

    # Select the subsets of delay and vectors to use
    delay_good = delay[pair_flags == Flag.GOOD]
    vectors_good = vectors_cart[pair_flags == Flag.GOOD]

    if method == 'jamaly':
        # Function that we'll minimize
        def vel_variance(theta, lag, vecs):
            # Project each vector into the direction of theta
            dir_dist = project_vectors(vecs, np.array(pol2rect(1, theta)))

            # Compute wind speed for each pair, ignoring infinites
            vel = dir_dist / lag
            vel[vel == np.inf] = np.nan
            vel[vel == -np.inf] = np.nan

            # compute the variance for this value of theta
            variance = np.nanvar(vel)
            return variance

        # Minimize the variance in the computed velocities to get theta. Use
        # only the Flag.Good values in delay and the vectors
        opt = minimize_scalar(vel_variance, args=(delay_good, vectors_good))
        cmv_theta = opt.x

        # Compute the velocity for all the useful pairs at this theta. The
        # actual velocity is the median over all useful pairs
        cmv_dir_dist = project_vectors(vectors_good,
                                       np.array(pol2rect(1, cmv_theta)))
        velocity = cmv_dir_dist / delay_good
        velocity[velocity == np.inf] = np.nan
        velocity[velocity == -np.inf] = np.nan
        cmv_vel = np.nanmedian(velocity)

        # QC - Compute the error index from Shinozaki et al (Jamaly eq 3)
        v60 = np.nanpercentile(velocity, 60)
        v40 = np.nanpercentile(velocity, 40)
        err = (v60 - v40) / cmv_vel
        if np.abs(err) > 0.4:  # Set flag if appropriate
            overall_flag = Flag.TREND

        method_out['v60'] = v60
        method_out['v40'] = v40
        method_out['error_index'] = err
        method_out['velocity'] = velocity

        # Compute a few extra statistics for the good pairs
        try:
            _, _, r_corr, _, std_err = linregress(delay_good, cmv_dir_dist)
        except ValueError:
            r_corr = 0
            std_err = 0
        ngood = len(vectors_good)
        method_out['r_corr'] = np.abs(r_corr)
        method_out['stderr_corr'] = std_err
        method_out['ngood'] = ngood

    elif method == 'gagne':
        # Function to apply least squares to
        def resid(p, lag, coeff_vecs):
            dx, dy = coeff_vecs[:, 0], coeff_vecs[:, 1]
            ax, ay = p[0], p[1]
            # Residual as defined in Gagne 2018 equation 3
            return lag - (dx * ax + dy * ay)

        [ax_opt, ay_opt], pcov = leastsq(resid, np.array([1, 1]),
                                         args=(delay_good, vectors_good))
        method_out['pcov'] = pcov

        # Gagne 2018 equation 4
        vx = 1 / (ax_opt ** 2 + ay_opt ** 2) * ax_opt
        vy = 1 / (ax_opt ** 2 + ay_opt ** 2) * ay_opt

        cmv_vel = magnitude([vx, vy])
        cmv_theta = np.arctan2(vy, vx)
    else:
        raise ValueError('Invalid method: ' + str(method) + '.')

    # Make velocity always positive and change distances to accommodate
    if cmv_vel < 0:
        cmv_vel *= -1
        cmv_theta += np.pi
    cmv_theta %= (2*np.pi)
    cmv_dir_dist = project_vectors(vectors_cart,
                                   np.array(pol2rect(1, cmv_theta)))

    # Create the output data object
    outdata = WindspeedData()
    outdata.flag = overall_flag
    outdata.pair_flag = pair_flags
    outdata.pair_lag = delay
    outdata.allpairs = pairs
    outdata.corr_lag = corr_lag
    outdata.pair_dists = cmv_dir_dist
    outdata.wind_angle = cmv_theta
    outdata.wind_speed = cmv_vel
    outdata.method_data = method_out

    return cmv_vel, cmv_theta, outdata


def optimum_subset(cmvx, cmvy, n=10):
    """
    Chooses a subset of vectors from within a full set of vectors, based on
    optimizing the most diverse angles available. Operates in 2 quadrants only,
    subject to the assumption that anti-parallels are also undesirable.

    This is useful for getting vectors that represent a variety of directions.
    For example, given all the CMVs for a whole year, this will provide a list
    of the time periods with the most varied cloud motion directions. One use
    case is in field analysis where the most diverse set of CMVs is needed to
    get the highest probability of perpendicular vector pairs.

    Parameters
    ----------
    cmvx : pd.Series
        A list of x-components of the CMVs

    cmvy : pd.Series
        A list of y-components of the CMVs

    n : int
        The number of vectors to select

    Returns
    -------
    indices : list
        The indices of the CMVs that are most diverse. Indices are positional
        within the input arrays. For example, could be used to reduce a
        DataFrame to the most diverse CMVs by using df.iloc[indices]
    """

    # Compute unit vectors representing the CMVs
    cld_spd = []
    for x, y in zip(cmvx, cmvy):
        cld_spd.append(magnitude((x, y)))
    cld_spd = np.array(cld_spd)
    cmv_x = cmvx / cld_spd
    cmv_y = cmvy / cld_spd
    cmv_vecs = np.array([cmv_x, cmv_y]).T

    def calc_cost(ang_0):
        """
        Compute the cost function associated with using each CMV as a member of
        the set relative to the ideal set of equally spaced vectors.

        Parameters
        ----------
        ang_0 : the rotation angle of the ideal vectors in radians

        Returns
        -------
        cost : the total cost of the assignment
        indices : the indices of the CMVs in the optimal assignment
        """
        # Compute unit vectors equally distributed around 180 deg.
        ideal_angs = np.arange(0, n) / n * np.pi + ang_0
        ideal_vecs = np.array([np.cos(ideal_angs), np.sin(ideal_angs)]).T

        # Compute cost as dot products between each CMV and each ideal vector
        # Absolute value used because both parallel and anti-parallel are bad
        # for our case.
        dots = -np.array([np.abs(dot(cmv_vecs.T, ideal_vec))
                          for ideal_vec in ideal_vecs])

        # Compute the optimal assignment of CMVs to maximize alignment with the
        # ideal vectors. Relative cost could be used to compare multiple zero
        # angles.
        r_ind, c_ind = linear_sum_assignment(dots)
        cost = dots[r_ind, c_ind].sum()
        return cost, c_ind

    def cost_wrapper(ang_0):
        """
        The minimizer only works on a single output, so in this case we wrap it
        """
        return calc_cost(ang_0)[0]  # return only the cost

    # The bounds of the optimization are limited by the spacing between ideal
    # vectors.
    bounds = [(0, np.pi / n)]

    #  What we're optimizing here is a rotation angle for the set of ideal vecs
    y = shgo(cost_wrapper, bounds)

    # Use the optimized angle to figure out which CMVs to use
    final_cost, indices = calc_cost(y.x[0])

    return indices


def _pairwise_qc(pairs, spacing, sigs, delay, peak_corr, mean_corr,
                 method, options=None):
    """
    Perform pairwise QC on the CMV signals according to method

    Parameters
    ----------
    pairs : np.array
        An N x 2 numpy array containing all the combination pairs.
    spacing : np.array
        An array of the distances between each pair in meters
    sigs : list of np.array
        A list of the two signals to be compared
    delay : np.array
        An array of the delays between each pair in seconds
    peak_corr : np.array
        An array of the peak cross-correlation between each pair
    mean_corr : np.array
        An array of the mean cross-correlation between each pair
    method : str
        The method to use for QC. Currently accepted methods are 'jamaly' and
        'gagne'
    options : dict or None
        A dictionary of options for the QC. See _validate_method_options for
        details
    """
    # Validate the options
    options, method_out = _validate_method_options(method, options)

    A, B = sigs

    # Initialize the flags for all the pairs
    npts = A.shape[1]
    pair_flags = np.empty(npts, dtype=Flag)
    pair_flags.fill(Flag.GOOD)

    # Global QC Common to all Methods

    # Same sensor lag and correlation is meaningless
    same_ind = pairs[:, 0] == pairs[:, 1]
    pair_flags[same_ind] = Flag.SAME

    # Base STD rejects faulty sensors with no variance at all, possibly bad
    # sensor data?
    siginds = np.bitwise_or(np.nanstd(A, axis=0) < 0.001,
                            np.nanstd(B, axis=0) < 0.001)
    pair_flags[siginds] = Flag.NOSIGNAL

    # Zero lag Reject sensor pairs with no delay
    pair_flags[delay == 0] = Flag.NOLAG

    # Signals are extremely overcast
    ktlim = options['ktlim']
    ktinds = np.bitwise_or(np.nanmax(A, axis=0) < ktlim,
                           np.nanmax(B, axis=0) < ktlim)
    pair_flags[ktinds] = Flag.OVERCAST

    # Method-specific QC
    if method == 'jamaly':
        # Jamaly and Kleissl - low variation ratio
        var_s_lim = options['var_s_lim']
        var_s0 = 1 - np.nanmean(A, axis=0) / np.nanmax(A, axis=0)
        var_s1 = 1 - np.nanmean(B, axis=0) / np.nanmax(B, axis=0)
        method_out['var_s'] = [var_s0, var_s1]
        var_s_inds = np.bitwise_or(var_s0 < var_s_lim, var_s1 < var_s_lim)
        pair_flags[var_s_inds] = Flag.LOWVAR_S

        # # Jamaly and Kleissl - low cloud cover fraction (i.e. too clear)
        # ccf_lim = 0.1  # More than 10% of points should have kt < 0.85
        # ccf0 = np.nansum(sig0 < 0.85) / len(sig0)  # Fraction of kt<0.85
        # ccf1 = np.nansum(sig1 < 0.85) / len(sig1)
        # if ccf0 < ccf_lim or ccf1 < ccf_lim:
        #     pair_flags[pair_ind] = Flag.LOWCCF
        #     continue

        # Jamaly and Kleissl - require minimum lagged xcorr and minimum
        # cross correlation ratio
        r_qc = 1 - mean_corr / peak_corr
        corr_min = options['mincorr']
        method_out['r_qc'] = r_qc
        corr_inds = np.bitwise_or(peak_corr < corr_min, r_qc < corr_min)
        pair_flags[corr_inds] = Flag.LOWCORR

        # Jamaly and Kleissl distance limit is confusing; I modified here
        # Original appears to require that the max velocity be less than
        # 1 m/s I implement it as a literal velocity limit rather than dist
        min_v = options['minvelocity']  # m/s (could be introduced)
        max_v = options['maxvelocity']  # m/s (70 m/s, around 160 mph)
        v = spacing/delay
        vind = np.bitwise_not(np.bitwise_and(min_v < np.abs(v),
                                             np.abs(v) < max_v))
        pair_flags[vind] = Flag.VELLIMIT

    elif method == 'gagne':
        pass  # No pairwise QC

    return pair_flags, method_out


def _validate_method_options(method, options):
    """
    Validate and fill defaults for an options dictionary for the various CMV
    methods.

    Parameters
    ----------
    method : str
        The method to validate options for. Currently accepted methods are:
        'jamaly' and 'gagne'.
    options : dict
        The options dictionary to validate. If None, a default dictionary will
        be created.

    Returns
    -------
    options : dict
        The validated options dictionary filled with defaults as applicable.
    method_out : dict
        A dictionary of method-specific output data. This is used to store
        intermediate data for debugging purposes.
    """
    # Validate options
    if options is None:
        options = {}
    # Specify default options by method
    if method == 'jamaly':
        defaults = {'minvelocity': 0,  # m/s
                    'maxvelocity': 70,  # m/s (about 160 mph)
                    'var_s_lim': 0.05,  # 0.10 in original paper
                    'ktlim': 0.4,
                    'mincorr': 0.8
                    }
        method_out = {
            'error_index': None,  # The Error Index
            'velocity': None,  # Velocity for each pair
            'v60': None,  # 60th percentile velocity
            'v40': None,  # 40th percentile velocity
            'r_qc': [],  # Ratio of peak to mean xcorr
            'var_s': [],  # Variation ratio of signals [s0, s1]
            'r_corr': None,  # Corr. coeff. for dist/delay for good pairs
            'stderr_corr': None,  # Std. error for dist/delay for good pairs
            'ngood': None,  # Number of good pairs
        }
    elif method == 'gagne':
        defaults = {
            'ktlim': 0.4,
        }
        method_out = {
            'pcov': None,  # Covariance matrix from least squares
        }
    else:
        raise ValueError('Invalid method: ' + str(method) + '.')

    # Validate option keys are valid for method
    for key in options:
        if key not in defaults:
            raise ValueError(f'Invalid option for {method}: ' + str(key) + '.')

    # Apply defaults if option not specified
    for key in defaults:
        if key not in options:
            options[key] = defaults[key]

    return options, method_out



from pytest import approx, fixture, raises

import numpy as np
import pandas as pd

from solarspatialtools import spatial


@fixture
def olat():
    # Oldenburg latitude
    olat = 53.143452
    return olat


@fixture
def olon():
    # Oldenburg longitude
    olon = 8.214571
    return olon


@fixture
def dlat():
    # These shifts are only approximately 1 40m cell distance,
    # so we assert relative to a buffer
    dlat = 3.64e-4
    return dlat


@fixture
def dlon():
    # These shifts are only approximately 1 40m cell distance,
    # so we assert relative to a buffer
    dlon = 6e-4
    return dlon


@fixture
def xypts():
    ys = [0, 0, 0, 1, -1, 1, -4]
    xs = [9, 8, 10, 9, 9, 10, 12]
    return xs, ys


@fixture
def projected_dxdy():
    expect = [[0, 0],
              [-1, 0],
              [1, 0],
              [0, 1],
              [0, -1],
              [1, 1],
              [3, -4]]
    return expect


@fixture
def projected_dists():
    c45 = np.cos(np.deg2rad(45))
    expect = {(1, 0): [0, -1, 1, 0, 0, 1, 3],
              (0, 1): [0, 0, 0, 1, -1, 1, -4],
              (1, 1): [0, -c45, c45, c45, -c45, 2*c45,
                       3*c45 - 4*c45]}
    return expect


def test__calc_zone():
    # Outer limits
    assert spatial._calc_zone(-179) == approx(1)
    assert spatial._calc_zone(-180) == approx(1)
    assert spatial._calc_zone(179) == approx(60)
    assert spatial._calc_zone(180) == approx(60)

    # Boundaries of Zone 32
    assert spatial._calc_zone(6) == approx(32)
    assert spatial._calc_zone(9) == approx(32)
    assert spatial._calc_zone(12) == approx(33)


def test_latlon2utm_sitecompare(olat, olon):
    # Value compared with MATLAB known good numbers
    (E, N, zone) = spatial.latlon2utm(olat, olon)
    assert (E, N, zone) == approx((447464.9005, 5888516.6988, 32))


def test_latlon2utm_failure(olat, olon):
    latarray = np.array([olat, olat, olat])
    lonarray = np.array([olon, olon, olon])
    with raises(ValueError):
        (E, N, zone) = spatial.latlon2utm(latarray, lonarray[0:2])
    with raises(ValueError):
        (E, N, zone) = spatial.latlon2utm(latarray[0:2], lonarray)


def test_latlon2utm_arrayin(olat, olon):
    # Value compared with MATLAB known good numbers
    latarray = np.array([olat, olat, olat])
    lonarray = np.array([olon, olon, olon])
    enlist = spatial.latlon2utm(latarray, lonarray)
    for EN in enlist:
        assert(EN[0], EN[1]) == approx((447464.9005, 5888516.6988))


def test_latlon2utm_pandas(olat, olon):
    latarray = np.array([olat, olat, olat])
    lonarray = np.array([olon, olon, olon])
    ind = [1, 2, 3]
    latseries = pd.Series(latarray, index=ind)
    lonseries = pd.Series(lonarray, index=ind)
    enframe = spatial.latlon2utm(latseries, lonseries)
    assert isinstance(enframe, pd.DataFrame)
    assert (enframe.columns == ['E', 'N', 'zone']).all()
    assert (enframe.index == ind).all()
    for enval in np.array(enframe):
        assert (enval[0], enval[1], enval[2]) \
               == approx((447464.9005, 5888516.6988, 32))


def test_latlon2utm_theory():
    # Test theoretically known point for zone 32
    (E, N, zone) = spatial.latlon2utm(0, 9)
    assert (E, N, zone) == approx((500000, 0, 32))

    # Test theoretically known point for zone 1
    (E, N, zone) = spatial.latlon2utm(0, -177)
    assert (E, N, zone) == approx((500000, 0, 1))

    # Test theoretically known point for zone 32
    (E, N, zone) = spatial.latlon2utm(0, 9, zone=32)
    assert (E, N, zone) == approx((500000, 0, 32))

    # Test theoretically known point for zone 1
    (E, N, zone) = spatial.latlon2utm(0, -177, zone=1)
    assert (E, N, zone) == approx((500000, 0, 1))


def test_latlon2utm_32north_defn():
    # Coords from https://epsg.io/32632

    # Bottom Left Corner (westmost extent)
    (E, N, zone) = spatial.latlon2utm(0, 6)
    assert (E, N, zone) == approx((166021.44, 0, 32))

    # Center
    (E, N, zone) = spatial.latlon2utm(84 / 2, 9)
    assert (E, N, zone) == approx((500000, 4649776.22, 32))

    # Top Right Corner (also northmost extent)
    (E, N, zone) = spatial.latlon2utm(84, 12 - 1e-10)
    assert (E, N, zone) == approx((534994.66, 9329005.18, 32))

    # Top Left Corner (should also be northmost extent)
    (E, N, _) = spatial.latlon2utm(84, 6)
    assert N == approx(9329005.18)


def test_latlon2utm_32south_defn():
    # Test for the southern hemisphere
    # Coords from https://epsg.io/32732

    # Bottom Left Corner
    (E, N, zone) = spatial.latlon2utm(-80, 6)
    assert (E, N, zone) == approx((441867.78, 1116915.04, 32))

    # Center
    (E, N, zone) = spatial.latlon2utm(-80 / 2, 9)
    assert (E, N, zone) == approx((500000, 5572242.78, 32))

    # Top Right Corner
    (E, N, zone) = spatial.latlon2utm(0 - 1e-12, 12 - 1e-10)
    assert (E, N, zone) == approx((833978.56, 10000000.00, 32))

    # Bottom Right Corner (should also be southmost extent)
    (E, N, _) = spatial.latlon2utm(-80, 12 - 1e-10)
    assert N == approx(1116915.04)


def test_utm2latlon_sitecompare(olat, olon):
    # Value compared with MATLAB known good numbers
    (lat, lon) = spatial.utm2latlon(447464.9005, 5888516.6988, zone=32)
    assert (lat, lon) == approx((olat, olon))


def test_utm2latlon_failure(olat, olon):
    earray = np.array([447464.9005, 447464.9005, 447464.9005])
    narray = np.array([5888516.6988, 5888516.6988, 5888516.6988])
    with raises(ValueError):
        spatial.utm2latlon(earray, narray[0:2], zone=32)
    with raises(ValueError):
        spatial.utm2latlon(earray[0:2], narray, zone=32)


def test_utm2latlon_arrayin(olat, olon):
    # Value compared with MATLAB known good numbers
    earray = np.array([447464.9005, 447464.9005, 447464.9005])
    narray = np.array([5888516.6988, 5888516.6988, 5888516.6988])
    latlonlist = spatial.utm2latlon(earray, narray, zone=32)
    for latlon in latlonlist:
        assert(latlon[0], latlon[1]) == approx((olat, olon))


def test_utm2latlon_pandas(olat, olon):
    # Value compared with MATLAB known good numbers
    earray = np.array([447464.9005, 447464.9005, 447464.9005])
    narray = np.array([5888516.6988, 5888516.6988, 5888516.6988])
    ind = [1, 2, 3]
    eseries = pd.Series(earray, index=ind)
    nseries = pd.Series(narray, index=ind)
    latlonlist = spatial.utm2latlon(eseries, nseries, zone=32)
    assert isinstance(latlonlist, pd.DataFrame)
    assert (latlonlist.columns == ['lat', 'lon']).all()
    assert (latlonlist.index == ind).all()
    for latlon in np.array(latlonlist):
        assert (latlon[0], latlon[1]) == approx((olat, olon))


def test_utm2latlon_theory():
    # Test theoretically known point for zone 32
    (lat, lon) = spatial.utm2latlon(500000, 0, zone=32)
    assert (lat, lon) == approx((0, 9))

    # Test theoretically known point for zone 1
    (lat, lon) = spatial.utm2latlon(500000, 0, zone=1)
    assert (lat, lon) == approx((0, -177))


def test_utm2latlon_32north_defn():
    # Coords from https://epsg.io/32632

    # Bottom Left Corner (westmost extent)
    (lat, lon) = spatial.utm2latlon(166021.44, 0, 32)
    assert (lat, lon) == approx((0, 6))

    # Center
    (lat, lon) = spatial.utm2latlon(500000, 4649776.22, 32)
    assert (lat, lon) == approx((84/2, 9))

    # Top Right Corner (also northmost extent)
    (lat, lon) = spatial.utm2latlon(534994.66, 9329005.18, 32)
    assert (lat, lon) == approx((84, 12 - 1e-10))


def test_utm2latlon_32south_defn():
    # Test for the southern hemisphere
    # Coords from https://epsg.io/32732

    # Bottom Left Corner
    (lat, lon) = spatial.utm2latlon(441867.78, 1116915.04, 32, True)
    assert (lat, lon) == approx((-80, 6))

    # Center
    (lat, lon) = spatial.utm2latlon(500000, 5572242.78, 32, True)
    assert (lat, lon) == approx((-80/2, 9))

    # Top Right Corner
    (lat, lon) = spatial.utm2latlon(833978.56, 10000000.00, 32, True)
    assert (lat, lon) == approx((0 - 1e-12, 12 - 1e-10))


def test_latlon2lcs_origin(olat, olon):
    (E, N) = spatial.latlon2lcs(olat, olon, olat, olon)
    assert (E, N) == approx((0, 0))


def test_latlon2lcs_shift(olat, olon, dlat, dlon):
    (E, N) = spatial.latlon2lcs(olat + dlat, olon + dlon, olat, olon)
    assert 39 < E < 41
    assert 39 < N < 41

    (E, N) = spatial.latlon2lcs(olat - dlat, olon - dlon, olat, olon)
    assert -41 < E < -39
    assert -41 < N < -39

    (E, N) = spatial.latlon2lcs(olat + dlat, olon + dlon, olat, olon, zone=32)
    assert 39 < E < 41
    assert 39 < N < 41

    (E, N) = spatial.latlon2lcs(olat - dlat, olon - dlon, olat, olon, zone=32)
    assert -41 < E < -39
    assert -41 < N < -39


def test_lcs2latlon_origin(olat, olon):
    lat, lon = spatial.lcs2latlon(0, 0, olat, olon)
    assert (lat, lon) == approx((olat, olon))


def test_lcs2latlon_shift(olat, olon, dlat, dlon):
    lat, lon = spatial.lcs2latlon(40, 40, olat, olon)
    assert (lat, lon) == approx((olat + dlat, olon + dlon), rel=1e-5)

    lat, lon = spatial.lcs2latlon(-40, -40, olat, olon)
    assert (lat, lon) == approx((olat - dlat, olon - dlon), rel=1e-5)

    # Repeat with specified zone
    lat, lon = spatial.lcs2latlon(40, 40, olat, olon, 32)
    assert (lat, lon) == approx((olat + dlat, olon + dlon), rel=1e-5)

    lat, lon = spatial.lcs2latlon(-40, -40, olat, olon, 32)
    assert (lat, lon) == approx((olat - dlat, olon - dlon), rel=1e-5)


def test_dot_list():
    dp = spatial.dot([1, 1], [-1, 1])
    assert dp == approx(0)

    dp = spatial.dot([1, 1], [1, 1])
    assert dp == approx(2)


def test_dot_numpy():
    dp = spatial.dot(np.array([1, 1]), np.array([-1, 1]))
    assert dp == approx(0)

    dp = spatial.dot(np.array([1, 1]), np.array([1, 1]))
    assert dp == approx(2)


def test_unit_list():
    u1 = spatial.unit([1, 0])
    assert u1 == approx(np.array([1, 0]))
    u2 = spatial.unit([-1, 1])
    assert u2 == approx(np.array([-1/np.sqrt(2), 1/np.sqrt(2)]))


def test_unit_numpy():
    u1 = spatial.unit(np.array([1, 0]))
    assert u1 == approx(np.array([1, 0]))
    u2 = spatial.unit(np.array([-1, 1]))
    assert u2 == approx(np.array([-1/np.sqrt(2), 1/np.sqrt(2)]))


def test_magnitude():
    assert spatial.magnitude([1, 0]) == approx(1)
    assert spatial.magnitude([1, 1]) == approx(np.sqrt(2))
    assert spatial.magnitude(np.array([1, 1])) == approx(np.sqrt(2))


def test_rect2pol():
    v1 = spatial.rect2pol(1, 0)
    assert v1 == approx(np.array([1, 0]))

    v1 = spatial.rect2pol(0, 1)
    assert v1 == approx(np.array([1, np.pi / 2]))

    v1 = spatial.rect2pol(1, 1)
    assert v1 == approx(np.array([np.sqrt(2), np.pi / 4]))

    v1 = spatial.rect2pol(1, -1)
    assert v1 == approx(np.array([np.sqrt(2), -np.pi / 4]))


def test_pol2rect():
    v1 = spatial.pol2rect(1, 0)
    assert v1 == approx(np.array([1, 0]))

    v1 = spatial.pol2rect(1, np.pi / 2)
    assert v1 == approx(np.array([0, 1]))

    v1 = spatial.pol2rect(np.sqrt(2), np.pi / 4)
    assert v1 == approx(np.array([1, 1]))

    v1 = spatial.pol2rect(np.sqrt(2), -np.pi / 4)
    assert v1 == approx(np.array([1, -1]))


def test_rotate_vector_list():
    r1 = spatial.rotate_vector([1, 0], np.deg2rad(90))
    assert r1 == approx([0, 1])

    r1 = spatial.rotate_vector([1, 0], np.deg2rad(-90))
    assert r1 == approx([0, -1])

    r1 = spatial.rotate_vector([1, 1], np.deg2rad(90))
    assert r1 == approx([-1, 1])

    r1 = spatial.rotate_vector([1, 1], np.deg2rad(-90))
    assert r1 == approx([1, -1])

    r1 = spatial.rotate_vector([1, 1], np.deg2rad(45))
    assert r1 == approx([0, np.sqrt(2)])


def test_rotate_vector_numpy():
    r1 = spatial.rotate_vector(np.array([1, 0]), np.deg2rad(90))
    assert r1 == approx([0, 1])

    r1 = spatial.rotate_vector(np.array([1, 0]), np.deg2rad(-90))
    assert r1 == approx([0, -1])

    r1 = spatial.rotate_vector(np.array([1, 1]), np.deg2rad(90))
    assert r1 == approx([-1, 1])

    r1 = spatial.rotate_vector(np.array([1, 1]), np.deg2rad(-90))
    assert r1 == approx([1, -1])

    r1 = spatial.rotate_vector(np.array([1, 1]), np.deg2rad(45))
    assert r1 == approx([0, np.sqrt(2)])


def test_rotate_vector_multiple():
    r1 = spatial.rotate_vector(np.array([[1, 0, 1], [0, 1, 1]]), np.deg2rad(90))
    assert r1 == approx(np.array([[0, -1, -1], [1, 0, 1]]))


def test_project_points(projected_dxdy, projected_dists):
    vecs = projected_dists.keys()
    for vec in vecs:
        dists = spatial.project_vectors(projected_dxdy, vec)
        assert dists == approx(projected_dists[vec])


def test_project_points_pandas(projected_dxdy, projected_dists):
    vecs = projected_dists.keys()
    for vec in vecs:
        ind = np.arange(len(projected_dxdy))
        frame = pd.DataFrame(projected_dxdy, index=ind)
        dists = spatial.project_vectors(frame, vec)
        assert isinstance(dists, pd.DataFrame)
        assert (dists.columns == ['dist']).all()
        assert (dists.index == ind).all()
        assert np.array(dists) == approx(projected_dists[vec])


def test_compute_vectors(xypts, projected_dxdy):
    xs = xypts[0]
    ys = xypts[1]
    refpt = (xs[0], ys[0])
    vecs = spatial.compute_vectors(xs, ys, refpt)
    for veca, vecb in zip(vecs, projected_dxdy):
        assert veca == approx(vecb)


def test_compute_vectors_pandas(xypts, projected_dxdy):
    xs = xypts[0]
    ys = xypts[1]
    refpt = (xs[0], ys[0])
    ind = np.arange(len(xs))
    vecs = spatial.compute_vectors(pd.Series(xs, index=ind),
                                   pd.Series(ys, index=ind),
                                   refpt)
    assert isinstance(vecs, pd.DataFrame)
    assert (vecs.columns == ['dx', 'dy']).all()
    assert (vecs.index == ind).all()
    for veca, vecb in zip(np.array(vecs), projected_dxdy):
        assert veca == approx(vecb)


@fixture(params=range(9))
def data(request):
    # [   (A),     (B),       (C)]
    # [(Ax, Ay), (Bx, By)), (Cx, Cy)]
    srcdata = [
        [(1, 0), (0, 1),
         (1, 1)],
        # initial rotated -45 degrees
        [(1/np.sqrt(2), 1/np.sqrt(2)), (1/np.sqrt(2), -1/np.sqrt(2)),
         (np.sqrt(2), 0)],
        # initial rotated +135 degrees
        [(-1/np.sqrt(2), -1/np.sqrt(2)), (-1/np.sqrt(2), 1/np.sqrt(2)),
         (-np.sqrt(2), 0)],
        # initial case rotated -60 degrees
        [(np.sqrt(3)/2, 1/2), (1/2, -np.sqrt(3)/2),
         (np.sqrt(2)*np.cos(15*np.pi/180), -np.sqrt(2)*np.sin(15*np.pi/180))],
        # Different lengths - Longer X
        [(4, 0), (0, 1),
         (4, 1)],
        # Longer Y
        [(1, 0), (0, 4),
         (1, 4)],
        # Longer X & Y
        [(3, 0), (0, 4),
         (3, 4)],
        # 45 degrees separation
        [(np.sqrt(2), 0), (1/np.sqrt(2), 1/np.sqrt(2)),
         (np.sqrt(2), 0)],
        # 60 degree separation evenly about x axis
        [(np.sqrt(3), 1), (np.sqrt(3), -1),
         (4/np.sqrt(3), 0)],
    ]
    return srcdata[request.param]


@fixture(params=[1, 10])
def n(request):
    return request.param


def test_compute_intersection(data, n):

    A = data[0]
    B = data[1]
    expected = data[2]

    A = np.repeat(np.array([A]), n, axis=0)
    B = np.repeat(np.array([B]), n, axis=0)
    expected = np.repeat(np.array([expected]), n, axis=0)

    C = spatial.compute_intersection(A, B)

    assert C == approx(expected)



import pytest
from pytest import approx, raises

import numpy as np
import numpy.testing as npt
import pandas as pd

import solarspatialtools.signalproc
from solarspatialtools import stats


class TestErrorFunctions:
    """
    rmse - root mean square error
    mse - mean square error
    se - squared error
    mae - mean absolute error
    ae - absolute error
    mbe - mean bias error
    be - bias error
    """

    @pytest.fixture
    def datax(self):
        datax = np.arange(0, 10)  # baseline
        return datax

    @pytest.fixture
    def datay(self):
        datay = np.arange(2, 12)  # baseline + 2
        return datay

    def test_rmse_identity(self, datax):
        assert stats.rmse(datax, datax) == approx(0)

    def test_rmse_value(self, datax, datay):
        assert stats.rmse(datax, datay) == approx(2)
        assert stats.rmse(datay, datax) == approx(2)

    def test_mse_identity(self, datax):
        assert stats.mse(datax, datax) == approx(0)

    def test_mse_value(self, datax, datay):
        assert stats.mse(datax, datay) == approx(4)
        assert stats.mse(datay, datax) == approx(4)

    def test_se_identity(self, datax):
        assert stats.squared_error(datax, datax).sum() == approx(0)

    def test_se_value(self, datax, datay):
        assert stats.squared_error(datax, datay).sum() == approx(40)
        assert stats.squared_error(datay, datax).sum() == approx(40)

    def test_mae_identity(self, datax):
        assert stats.mae(datax, datax).sum() == approx(0)

    def test_mae_value(self, datax, datay):
        assert stats.mae(datax, datay).sum() == approx(2)
        assert stats.mae(datay, datax).sum() == approx(2)

    def test_ae_identity(self, datax):
        assert stats.absolute_error(datax, datax).sum() == approx(0)

    def test_ae_value(self, datax, datay):
        assert stats.absolute_error(datax, datay).sum() == approx(20)
        assert stats.absolute_error(datay, datax).sum() == approx(20)

    def test_mbe_identity(self, datax):
        assert stats.mbe(datax, datax).sum() == approx(0)

    def test_mbe_value(self, datax, datay):
        assert stats.mbe(datax, datay).sum() == approx(2)
        assert stats.mbe(datay, datax).sum() == approx(-2)

    def test_be_identity(self, datax):
        assert stats.bias_error(datax, datax).sum() == approx(0)

    def test_be_value(self, datax, datay):
        assert stats.bias_error(datax, datay).sum() == approx(20)
        assert stats.bias_error(datay, datax).sum() == approx(-20)


class TestQuantile:
    """
    Computations of the quantile level for a time series based on prior
    history of days
    """

    @pytest.fixture
    def quantile_data(self):
        """
        Four days of data with a constant value
        """
        index = pd.date_range(start='2019-04-22 00:00:00',
                              end='2019-04-26 23:59:59', freq='1h')
        data = pd.DataFrame(5, index=index, columns=["ColName"])
        return data

    def test_calc_quantile(self, quantile_data):
        data = quantile_data
        p90 = stats.calc_quantile(data, n_days='2d', quantile=0.9)
        npt.assert_allclose(p90['ColName_quant']['2019-04-23 23:00:00':
                                                 '2019-04-24 00:00:00'],
                            [np.nan, 5])

    def test_calc_quantile_series(self, quantile_data):
        data = quantile_data
        p90 = stats.calc_quantile(data['ColName'], n_days='2d', quantile=0.9)
        npt.assert_allclose(p90['ColName_quant']['2019-04-23 23:00:00':
                                                 '2019-04-24 00:00:00'],
                            [np.nan, 5])

    def test_calc_quantile_frame(self, quantile_data):
        data = quantile_data
        dat = np.ones([len(data), 2])
        data2d = pd.DataFrame(dat, index=data.index,
                              columns=["ColName", "Another"])
        with pytest.raises(ValueError):
            stats.calc_quantile(data2d, n_days='2d', quantile=0.9)


class TestVariabilityMetrics:
    @pytest.fixture
    def ghi(self):
        index = pd.date_range(start='2019-04-22 00:00:00',
                              end='2019-04-22 00:00:05', freq='1s')
        ghi = pd.Series([1, 2, 4, 7, 11, 16], index=index)
        return ghi

    @pytest.fixture
    def darr_data(self):
        index = pd.date_range(start='2019-04-22 00:00:00',
                              end='2019-04-26 02:59:59', freq='1min')
        data = pd.DataFrame(5, index=index, columns=["ColName"])
        return data

    @pytest.fixture
    def variability_score_data(self):
        index = pd.date_range(start='2019-04-22 00:00:00',
                              end='2019-04-26 02:59:59', freq='1min')
        data = pd.DataFrame(5, index=index, columns=["ColName"])
        cs = pd.DataFrame(5, index=index, columns=["ColName"])
        data.iloc[5] = 6
        return data, cs

    def test_variability_index_null(self, variability_score_data):
        _, cs = variability_score_data
        vi = stats.variability_index(cs, cs, moving_avg_tau=1, norm=False)
        assert vi.values == approx(1)

    def test_variability_index_basic(self, variability_score_data):
        data, cs = variability_score_data
        vi = stats.variability_index(data, cs, moving_avg_tau=1, norm=False)
        assert vi.values == approx((5937 + 2 * np.sqrt(2)) / 5939)

    def test_variability_index_illegal(self, variability_score_data):
        data, cs = variability_score_data
        with raises(TypeError):
            stats.variability_index(data.values, cs)

    def test_variability_index_movingavg(self, variability_score_data):
        data, cs = variability_score_data
        vi = stats.variability_index(data, cs, moving_avg_tau=2, norm=False)
        assert vi.values == approx((2 * 2967 + 2 * np.sqrt(2 ** 2 + 0.5 ** 2))
                                    / (2 * 2969))

    @pytest.fixture
    def variability_score_data_multi(self):
        index = pd.date_range(start='2019-04-22 00:00:00',
                              end='2019-04-26 02:59:59', freq='1min')
        dat = 5 * np.ones([len(index), 3])
        cols = ["ColName1", "ColName2", "ColName3"]
        data = pd.DataFrame(dat.copy(), index=index, columns=cols)
        cs = pd.DataFrame(dat.copy(), index=index, columns=cols)
        data.iloc[5, :] = 6
        return data, cs

    def test_variability_index_null_multi(self, variability_score_data_multi):
        _, cs = variability_score_data_multi
        vi = stats.variability_index(cs, cs, moving_avg_tau=1, norm=False)
        assert np.allclose(vi, 1)

    def test_variability_index_basic_multi(self, variability_score_data_multi):
        data, cs = variability_score_data_multi
        vi = stats.variability_index(data, cs, moving_avg_tau=1, norm=False)
        assert np.allclose(vi, (5937 + 2 * np.sqrt(2)) / 5939)

    def test_variability_index_movingavg_multi(self, variability_score_data_multi):
        data, cs = variability_score_data_multi
        vi = stats.variability_index(data, cs, moving_avg_tau=2, norm=False)
        assert np.allclose(vi, (2 * 2967 + 2 * np.sqrt(2 ** 2 + 0.5 ** 2))
                                / (2 * 2969))

    def test_darr_zero(self, darr_data):
        data = darr_data
        npt.assert_allclose(stats.darr(data, pct=False), 0)

    def test_darr_basic(self, darr_data):
        data = darr_data
        data.iloc[2] = 6

        assert (stats.darr(data, moving_avg=False, pct=False).values \
                == approx(2))
        assert stats.darr(data, pct=False).values == approx(2)

    def test_darr_illegal(self, darr_data):
        data = darr_data
        data.iloc[2] = 6
        with raises(TypeError):
            stats.darr(data.values, moving_avg=False, pct=False)

    def test_darr_pct(self, darr_data):
        data = darr_data
        data.iloc[2] = 6
        assert stats.darr(data, pct=True).values == approx(2 * 100 / 1000)

    def test_darr_tau(self, darr_data):
        data = darr_data
        data.iloc[2] = 6
        assert stats.darr(data, tau=2, moving_avg=False, pct=False).values \
               == approx(2)

    def test_darr_movingavg(self, darr_data):
        data = darr_data
        data.iloc[2] = 6
        assert (stats.darr(data, tau=2, moving_avg=True, pct=False).values \
                == approx(1))

    @pytest.fixture
    def darr_data_multi(self):
        index = pd.date_range(start='2019-04-22 00:00:00',
                              end='2019-04-26 02:59:59', freq='1min')
        dat = 5 * np.ones([len(index), 3])
        cols = ["ColName1", "ColName2", "ColName3"]
        data = pd.DataFrame(dat, index=index, columns=cols)
        return data

    def test_darr_zero_multi(self, darr_data_multi):
        data = darr_data_multi
        npt.assert_allclose(stats.darr(data, pct=False), 0)

    def test_darr_basic_multi(self, darr_data_multi):
        data = darr_data_multi
        data.iloc[2, :] = 6

        dar = stats.darr(data, moving_avg=False, pct=False)
        assert np.allclose(dar, 2)
        dar = stats.darr(data, pct=False)
        assert np.allclose(dar, 2)

    def test_darr_pct_multi(self, darr_data_multi):
        data = darr_data_multi
        data.iloc[2, :] = 6
        dar = stats.darr(data, pct=True)
        assert np.allclose(dar, 2 * 100 / 1000)

    def test_darr_tau_multi(self, darr_data_multi):
        data = darr_data_multi
        data.iloc[2, :] = 6
        dar = stats.darr(data, tau=2, moving_avg=False, pct=False)
        assert np.allclose(dar, 2)

    def test_darr_movingavg_multi(self, darr_data_multi):
        data = darr_data_multi
        data.iloc[2, :] = 6
        dar = stats.darr(data, tau=2, moving_avg=True, pct=False)
        assert np.allclose(dar, 1)

    def test_variability_score_basic(self, ghi):
        vs = stats.variability_score(ghi, tau=1, moving_avg=False, pct=False)
        assert vs == approx(1.2)  # 2 * 50% chance of being greater

    def test_variability_score_pandas(self, ghi):
        vs = stats.variability_score(pd.Series(ghi), tau=1,
                                     moving_avg=False, pct=False)
        assert vs == approx(1.2)  # 2 * 50% chance of being greater
        vs = stats.variability_score(pd.DataFrame(ghi), tau=1,
                                     moving_avg=False, pct=False)
        assert vs.values == approx(1.2)  # 2 * 50% chance of being greater

    def test_variability_score_pct(self, ghi):
        vs = stats.variability_score(100 * ghi, tau=1,
                                     moving_avg=False, pct=True)
        assert vs == approx(12)

    def test_variability_score_tau(self, ghi):
        vs = stats.variability_score(ghi, tau=2, moving_avg=False, pct=False)
        assert vs == approx(2.5)  # 3 * 66.667% chance of being greater

    def test_variability_score_movingavg(self, ghi):
        vs = stats.variability_score(ghi, tau=2, moving_avg=True, pct=False)
        assert vs == approx(2)  # 1.5 * 66.667% chance of being greater

    def test_variability_score_basic_multi(self, ghi):
        dfghi = pd.DataFrame({1: ghi, 2: ghi, 3: ghi}, index=ghi.index)
        vs = stats.variability_score(dfghi, tau=1, moving_avg=False, pct=False)
        assert np.allclose(vs, 1.2)
        assert vs.shape == (3,)

    def test_variability_score_pct_multi(self, ghi):
        dfghi = pd.DataFrame({1: ghi, 2: ghi, 3: ghi}, index=ghi.index)
        vs = stats.variability_score(100 * dfghi, tau=1,
                                     moving_avg=False, pct=True)
        assert np.allclose(vs, 12)
        assert vs.shape == (3,)

    def test_variability_score_tau_multi(self, ghi):
        dfghi = pd.DataFrame({1: ghi, 2: ghi, 3: ghi}, index=ghi.index)
        vs = stats.variability_score(dfghi, tau=2, moving_avg=False, pct=False)
        assert np.allclose(vs, 2.5)
        assert vs.shape == (3,)

    def test_variability_score_movingavg_multi(self, ghi):
        dfghi = pd.DataFrame({1: ghi, 2: ghi, 3: ghi}, index=ghi.index)
        vs = stats.variability_score(dfghi, tau=2, moving_avg=True, pct=False)
        assert np.allclose(vs, 2)
        assert vs.shape == (3,)



import numpy as np
import pandas as pd


def rmse(baseline, estimation):
    """
    Computation of Root Mean Squared Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric
        the root mean squared error
    """
    mse_val = mse(baseline, estimation)
    result = np.sqrt(mse_val)
    return result


def mse(baseline, estimation):
    """
    Computation of Mean Squared Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric
        the mean squared error
    """
    result = np.mean(squared_error(baseline, estimation))
    return result


def squared_error(baseline, estimation):
    """
    Time series of Squared Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric or pandas.Series
        a time series of squared error values
    """
    return bias_error(baseline, estimation) ** 2


def mae(baseline, estimation):
    """
    Computation of Mean Absolute Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric
        the mean absolute error
    """
    result = np.mean(absolute_error(baseline, estimation))
    return result


def absolute_error(baseline, estimation):
    """
    Time series of Absolute Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric or pandas.Series
        a time series of absolute error values
    """
    result = np.abs(bias_error(baseline, estimation))
    return result


def mbe(baseline, estimation):
    """
    Computation of Mean Bias Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric
        the mean bias error
    """
    result = np.mean(bias_error(baseline, estimation))
    return result


def bias_error(baseline, estimation):
    """
    Time series of Bias Error

    Parameters
    ----------
    baseline : numeric or pandas.Series
        Baseline time series

    estimation : numeric or pandas.Series
        Reference/Estimation time series

    Returns
    -------
    result : numeric or pandas.Series
        a time series of bias error values
    """
    result = estimation - baseline
    return result


def variability_score(series, tau=1, moving_avg=True, pct=False):
    """
    Compute the variability score as proposed by Lave et al. [1]. The
    Variability Score is computed by calculating the product of each ramp rate
    that occurs in the time series and the probability of occurrence of larger
    ramp rates than that ramp rate of interest. The maximum value of that
    product over all possible ramp rates is the Variability Score. Its value
    may be represented as a percent.

    The original source computes this quantity using the GHI, but we offer the
    possibility of computing it for clear sky index.

    Parameters
    ----------
    series : pandas.Series or pandas.DataFrame
        a time series for which to calculate the Variability Score. VS will
        be calculated along axis 0.

    tau : numeric, default 1
        The number of timesteps for the increment calculation. series must use
        a temporal index to use a dt greater than 1.

    moving_avg : bool, default True
        When tau specified with a value different from 1, should the timeseries
        be resampled via moving avg to the frequency of tau for computation?

    pct : bool, default False
        should we scale the score as a percent of 1000 W/m2 (0-100%)?

    Returns
    -------
    variability_score : numeric
        the variability score

    [1] M. Lave, M. J. Reno, and R. J. Broderick, “Characterizing local high-
    frequency solar variability and its impact to distribution studies,” Solar
    Energy 118, 327–337 (2015). https://www.osti.gov/pages/biblio/1497655
    """
    if moving_avg and tau > 1:
        dt = (series.index[1] - series.index[0]).total_seconds()
        rs = series.resample(str(int(tau * dt)) + 's').mean()
        return variability_score(rs, tau=1, moving_avg=False, pct=pct)
    if pct:  # Scale to a percentage of STC and multiply by 100 (Lave et al)
        return 100./1000 * variability_score(series, tau=tau,
                                             moving_avg=False, pct=False)

    x = series.diff(tau)
    vs = (x.abs() * (1. - x.abs().rank(pct=True))).max(axis=0)

    return vs


def variability_index(ghi, clearsky, moving_avg_tau=1, norm=False):
    """
    Compute the variability index, defined by Stein et al. [1]. Variability
    Index represents the ratio between the path length of the actual timeseries
    as compared to that of the clearsky.

    As it requires comparison with the clear sky value, this should only be
    computed using the GHI (or other irradiance), and not clear sky index.

    Parameters
    ----------
    ghi : pandas.Series or pd.DataFrame with only a single column
        a series containing the ghi

    clearsky : pandas.Series or pd.DataFrame with only a single column
        a series with the clear sky ghi. Must have a temporal index.

    moving_avg_tau : numeric
        The number of averaging timesteps to use

    norm : bool
        should the output be scaled to represent a dt of 1 minute?

    Returns
    -------
    variability_index : numeric
        the variability index value

    [1] J. S. Stein, C. W. Hansen, and M. J. Reno, “The Variability Index: A
    New and Novel Metric for Quantifying Irradiance and PV Output Variability,”
    in Proceedings of the World Renewable Energy Forum (Denver, CO, 2012)
    pp. 13–17. https://www.osti.gov/biblio/1078490
    """
    if not isinstance(ghi, (pd.Series, pd.DataFrame)):
        raise TypeError('series must be a pandas Series or DataFrame')

    if moving_avg_tau > 1:
        dti = (ghi.index[1] - ghi.index[0]).total_seconds()
        sample_pd = str(int(moving_avg_tau*dti))+'s'
        rs = ghi.resample(sample_pd).mean()
        cs = clearsky.resample(sample_pd).mean()
        return variability_index(rs, cs, moving_avg_tau=1, norm=norm)
    dt = (ghi.index[1] - ghi.index[0]).total_seconds() / 60
    num = np.sqrt(ghi.diff() ** 2 + dt ** 2).sum(axis=0)
    den = np.sqrt(clearsky.diff() ** 2 + dt ** 2).sum(axis=0)
    if norm:  # Stein reports VI_dt = VI_1min/sqrt(dt)
        vi = num/den * np.sqrt(dt)
    else:
        vi = num/den

    return vi

def darr(series, tau=1, moving_avg=True, pct=False):
    """
    Compute the Daily Aggregate Ramp Rate as described by Van Haaren et al. [1]

    The DARR represents the sum of all possible ramp rates, and according to
    the original source, could be normalized as a percent of STC irradiance.

    The original source computes this quantity using the GHI, but we offer the
    possibility of computing it for clear sky index.


    Parameters
    ----------
    series : pandas.Series
        a time series for which to calculate the Variability Score

    tau : numeric, default 1
        The number of timesteps for the increment calculation. series must use
        a temporal index to use a dt greater than 1.

    moving_avg : bool, default True
        When tau specified with a value different from 1, should the timeseries
        be resampled via moving avg to the frequency of tau for computation?

    pct : bool, default False
        should we scale the score as a percent of 1000 W/m2 (0-100%)?

    Returns
    -------
    darr : numeric
        The value of the metric

    [1] R. van Haaren, M. Morjaria, and V. Fthenakis, “Empirical assessment
    of short-term variability from utility-scale solar PV plants,” Progress in
    Photovoltaics: Research and Applications 22, 548–559 (2014),
    https://www.researchgate.net/publication/261603714_Empirical_assessment_of_short-term_variability_from_utility-scale_solar_PV_plants
    """
    if not isinstance(series, (pd.Series, pd.DataFrame)):
        raise TypeError('series must be a pandas Series or DataFrame')

    if moving_avg and tau > 1:
        dt = (series.index[1] - series.index[0]).total_seconds()
        rs = series.resample(str(int(tau * dt)) + 's').mean()
        # Rerun it without MA
        return darr(rs, tau=1, moving_avg=False, pct=pct)
    if pct:
        return darr(series * 100 / 1000, tau=tau, moving_avg=False, pct=False)

    darr_val = np.abs(series.diff(tau)).sum(axis=0)

    return darr_val


def calc_quantile(timeseries, n_days="30d", quantile=0.9):
    """
    Calculate a single-day percentile-based summary of data by aggregating
    multiple days.

    So the timeseries output on any given day represents the quantile-th
    percentile value over the past n-days during any particular timestamp.

    An example usage of this is in the QCPV algorithm of Killinger et al. [1],
    who use it on a time series of PV data to compute an empirical 'Clear Sky
    PV' value that can be used to simulate a PV clear sky index.

    Parameters
    ----------
    timeseries : pandas.Series or pandas.DataFrame
        the timeseries of input data. Needs to just be a single column or a
        Series

    n_days : str, default '30d'
        the number of days window to use for the calculation on quantile in a
        format readable by pandas.to_timedelta()

    quantile : numeric, default 0.9
        the quantile level as a fraction

    Returns
    -------
    output : pandas.DataFrame
        output timeseries. nan for first days of the series until enough days
        for the window to be met. The data will be stored in a column based on
        the input column name with "_quant" added.

    [1] Killinger, Sven & Engerer, Nick & Müller, Björn. (2017). QCPV: A
    quality control algorithm for distributed photovoltaic array power output.
    Solar Energy. 143. 120-131.
    https://www.researchgate.net/publication/312145487_QCPV_A_quality_control_algorithm_for_distributed_photovoltaic_array_power_output
    """

    ts = timeseries.copy()  # copy the data

    out_df = pd.Series()  # Create an empty holder

    end_day = ts.index.date[-1]  # the final day of the dataset

    # Loop over all unique days in the timeseries
    for date_start in pd.unique(ts.index.date):
        # Window starts on date_start. Window ends on date_end
        date_end = date_start + pd.to_timedelta(n_days)

        # stop when the window tries to reach beyond the end of the data
        if date_end > end_day:
            break

        # Copy the subset of data representing the window
        # and separate the date and time stamps
        statdata = ts[date_start:date_end]
        dates = statdata.index.date
        times = statdata.index.time

        # Build into a new DataFrame that can be pivoted
        statdata = pd.DataFrame({'data': statdata.values.flatten(), 'd': dates,
                                 't': times}, index=statdata.index)

        # Create a pivot table based on the time as the index so that we can
        # compute statistics across all the days
        piv = pd.pivot(statdata, index='t', columns='d', values='data')

        # Calculate the quantile
        p90 = piv.quantile(quantile, axis=1)
        p90.name = date_end  # Assign a name based on the end date

        # p90's index is only time, so modify it to include the full date
        p90.index = pd.to_datetime(np.repeat(str(date_end) + " ",
                                             len(p90.index)) +
                                   p90.index.astype(str))

        # Concat this day onto the output object
        if len(out_df) == 0:
            out_df = p90
        else:
            out_df = pd.concat((out_df, p90), axis=0)

    # localize the timeseries and the column names back to the input
    out_df = out_df.tz_localize(ts.index.tz).reindex(ts.index)
    out_df = pd.DataFrame(out_df)
    try:
        out_df.columns = [name+'_quant' for name in ts.columns]
    except AttributeError:
        try:
            out_df.columns = [ts.name + '_quant']
        except AttributeError:
            raise AttributeError('Unable to generate name of column')
        except TypeError:
            out_df.columns = ['quantile']

    return out_df



import pytest
from pytest import approx

import numpy as np
import pandas as pd

from solarspatialtools import cmv


@pytest.fixture(params=[0, 45, 60, -60, 90, 124.5, 340.5])
def theta_deg(request):
    return request.param


@pytest.fixture(params=[2, 5, 10, 30])
def velocity(request):
    return request.param


@pytest.fixture(params=['jamaly', 'gagne'])
def mode(request):
    return request.param


def test_cmv_artificial(theta_deg, velocity, mode):
    # Generate a dataframe with 9 points arranged in a 3x3 square grid
    # The points are arranged in a grid with 1m spacing
    # The origin is in the center of the grid
    # The points are numbered as follows:
    # 6 7 8
    # 3 4 5
    # 0 1 2
    # The reference point is 4
    # The CMV is 1m/s at 45°
    # The expected result is 1m/s at 45°
    pos_utm = pd.DataFrame(index=range(9), columns=['E', 'N'], dtype=float)
    spacing = 10.0
    pos_utm.loc[0] = [-spacing, -spacing]
    pos_utm.loc[1] = [0.0, -spacing]
    pos_utm.loc[2] = [spacing, -spacing]
    pos_utm.loc[3] = [-spacing, 0.0]
    pos_utm.loc[4] = [0.0, 0.0]
    pos_utm.loc[5] = [spacing, 0.0]
    pos_utm.loc[6] = [-spacing, spacing]
    pos_utm.loc[7] = [0.0, spacing]
    pos_utm.loc[8] = [spacing, spacing]

    # Velocity direction
    theta = theta_deg * np.pi/180.0

    # delays between points are deterministic
    delays = np.zeros(9)
    delays[4] = 0
    delays[5] = spacing*np.cos(theta)/velocity
    delays[7] = spacing*np.sin(theta)/velocity
    delays[3] = -spacing*np.cos(theta)/velocity
    delays[1] = -spacing*np.sin(theta)/velocity
    delays[8] = spacing*np.sqrt(2)*np.cos(theta-45*np.pi/180)/velocity
    delays[2] = -spacing*np.sqrt(2)*np.sin(theta-45*np.pi/180)/velocity
    delays[6] = spacing*np.sqrt(2)*np.sin(theta-45*np.pi/180)/velocity
    delays[0] = -spacing*np.sqrt(2)*np.cos(theta-45*np.pi/180)/velocity

    # Generate the base signal. High enough sample rate and T is required
    # to fully resolve the delays. We use noise to ensure broad frequency
    # content.
    fs = 500  # sample rate
    T = 500.0  # total seconds
    np.random.seed(2023)
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable
    noise = np.random.random(len(t))/5  # add broadband noise
    x = 0.5*np.sin(2*np.pi*2*t) + noise  # noisy signal

    # The individual signals are delayed versions of the original. Noise is
    # shifted rather than regenerated to ensure broadband coherence.
    signals = np.zeros((9, len(x)))
    signals[4, :] = x
    for i, delay in enumerate(delays):
        signals[i, :] = np.roll(x, int(delay*fs))
    df = pd.DataFrame(signals.T, index=pd.to_timedelta(t, 's'))

    # Compute the CMV
    cld_spd, cld_dir, dat = cmv.compute_cmv(df, pos_utm,
                                            reference_id=4,
                                            method=mode)

    # Convert the dir to positive angle, because compute_cmv is always pos
    if theta < 0:
        theta += 2*np.pi
    if theta > 2*np.pi:
        theta -= 2*np.pi

    assert (cld_spd == approx(velocity, rel=0.01))

    # Directions sometimes get wonky near the 0/360 boundary
    assert (cld_dir == approx(theta, rel=0.01) or
            cld_dir == approx(theta+2*np.pi, rel=0.01))


def test_cmv_gagne_data():
    datafile = "../demos/data/sample_plant_1.h5"
    pos_utm = pd.read_hdf(datafile, mode="r", key="utm")
    df = pd.read_hdf(datafile, mode="r", key="data_a")

    hourlymax = np.mean(df.quantile(0.95))
    kt = df / hourlymax

    cld_spd_gag, cld_dir_gag, dat_gag = cmv.compute_cmv(kt, pos_utm,
                                                        reference_id=None,
                                                        method='gagne')

    # Known values calculated on 11/3/2023
    assert cld_spd_gag == approx(10.53, abs=0.01)
    assert cld_dir_gag == approx(3.68, abs=0.01)
    assert sum(dat_gag.pair_flag == cmv.Flag.GOOD) == 574


def test_cmv_jamaly_data():
    datafile = "../demos/data/sample_plant_1.h5"
    pos_utm = pd.read_hdf(datafile, mode="r", key="utm")
    df = pd.read_hdf(datafile, mode="r", key="data_a")

    hourlymax = np.mean(df.quantile(0.95))
    kt = df / hourlymax

    cld_spd_jam, cld_dir_jam, dat_jam = cmv.compute_cmv(kt, pos_utm,
                                                        reference_id=None,
                                                        method='jamaly')

    assert cld_spd_jam == approx(10.54, abs=0.01)
    assert cld_dir_jam == approx(3.29, abs=0.01)
    assert sum(dat_jam.pair_flag == cmv.Flag.GOOD) == 12709

def test_cmv_jamaly_data_ref():
    # A test using a reference when we calculate the CMV

    datafile = "../demos/data/sample_plant_1.h5"
    pos_utm = pd.read_hdf(datafile, mode="r", key="utm")
    df = pd.read_hdf(datafile, mode="r", key="data_a")

    hourlymax = np.mean(df.quantile(0.95))
    kt = df / hourlymax

    cld_spd_jam, cld_dir_jam, dat_jam = cmv.compute_cmv(kt, pos_utm,
                                                        reference_id=pos_utm.index[0],
                                                        method='jamaly')


    # Known values calculated on 11/3/2023
    assert cld_spd_jam == approx(9.3911, abs=0.01)
    assert cld_dir_jam == approx(3.665955, abs=0.01)
    assert sum(dat_jam.pair_flag == cmv.Flag.GOOD) == 64


def test_optimum_subset_base():
    angles = np.linspace(0, np.pi, 9)
    magnitudes = np.linspace(1, 6, 9)
    vx = magnitudes * np.cos(angles)
    vy = magnitudes * np.sin(angles)
    indices = cmv.optimum_subset(vx, vy, n=4)
    assert np.sort(np.rad2deg(angles[indices])) == approx(np.sort(np.array([0, 45, 90, 135])))


def test_optimum_subset_rotate():
    angles = np.linspace(0, np.pi, 9) + np.pi/4
    magnitudes = np.linspace(1, 6, 9)
    vx = magnitudes * np.cos(angles)
    vy = magnitudes * np.sin(angles)
    indices = cmv.optimum_subset(vx, vy, n=4)
    assert np.sort(np.rad2deg(angles[indices])) == approx(np.sort(np.array([45, 90, 135, 180])))



import pytest
from pytest import approx

import numpy as np
import pandas as pd

from solarspatialtools import field, spatial


@pytest.fixture(params=['coherence', 'global_coherence', 'distance', 'all'])
def mode(request):
    return request.param


@pytest.fixture(params=[0, 1, 2, 3, 4])
def refdat(request, mode, delay_method):
    data = {}
    data['multi'] = {
        "coherence": (
            ('CMB-001-1', (46.35, 1117.45)),
            ('CMB-025-7', (237.82, 531.65)),
            ('CMB-015-3', (367.55, 811.45)),
            ('CMB-003-1', (244.10, 1039.52)),
            ('CMB-008-5', (164.93, 876.94))
        ),
        "global_coherence": (
            ('CMB-001-1', (53.37, 1107.52)),
            ('CMB-025-7', (234.23, 529.77)),
            ('CMB-015-3', (369.15, 809.23)),
            ('CMB-003-1', (245.82, 1054.97)),
            ('CMB-008-5', (164.31, 876.15))
        ),
        "distance": (
            ('CMB-001-1', (-36.97, 1063.08)),
            ('CMB-025-7', (239.94, 540.79)),
            ('CMB-015-3', (368.99, 810.51)),
            ('CMB-003-1', (258.80, 1067.81)),
            ('CMB-008-5', (162.10, 869.84))
        ),
        'all': (
            ('CMB-001-1', (348.09, 653.15)),
            ('CMB-025-7', (291.24, 542.36)),
            ('CMB-015-3', (414.61, 587.06)),
            ('CMB-003-1', (476.56, 652.45)),
            ('CMB-008-5', (359.54, 617.64))
        ),
    }

    data['fit'] = {
        "coherence": (
            ('CMB-001-1', (45.79, 1116.99)),
            ('CMB-025-7', (235.84, 531.57)),
            ('CMB-015-3', (368.11, 812.28)),
            ('CMB-003-1', (242.76, 1041.82)),
            ('CMB-008-5', (165.38, 876.23))
        ),
        "global_coherence": (
            ('CMB-001-1', (51.69, 1106.64)),
            ('CMB-025-7', (232.39, 528.94)),
            ('CMB-015-3', (369.86, 809.84)),
            ('CMB-003-1', (243.85, 1055.34)),
            ('CMB-008-5', (162.35, 874.83))
        ),
        "distance": (
            ('CMB-001-1', (-38.24, 1064.02)),
            ('CMB-025-7', (239.24, 540.23)),
            ('CMB-015-3', (371.59, 812.76)),
            ('CMB-003-1', (256.50, 1067.87)),
            ('CMB-008-5', (161.47, 868.25))
        ),
        'all': (
            ('CMB-001-1', (364.61, 611.05)),
            ('CMB-025-7', (293.98, 506.67)),
            ('CMB-015-3', (411.25, 586.00)),
            ('CMB-003-1', (474.20, 654.19)),
            ('CMB-008-5', (356.40, 613.11))
        ),
    }

    datai = data[delay_method][mode]

    return delay_method, mode, datai[request.param]


def test_compute_predicted_position_static(refdat):
    delay_method = refdat[0]
    mode = refdat[1]
    ref = refdat[2][0]
    expect = refdat[2][1]

    datafile = "../demos/data/sample_plant_2.h5"
    cmv_a = spatial.pol2rect(9.52, 0.62)
    cmv_b = spatial.pol2rect(8.47, 2.17)

    pos_utm = pd.read_hdf(datafile, mode="r", key="utm")

    df_a = pd.read_hdf(datafile, mode="r", key="data_a")
    df_b = pd.read_hdf(datafile, mode="r", key="data_b")

    pos, _ = field.compute_predicted_position(
        [df_a, df_b],  # The dataframes with the two one hour periods
        pos_utm,  # the dataframe specifying the combiner positions
        ref,  # the position within pos_utm to calculate about
        [cmv_a, cmv_b],  # The two individual CMVs for the DFs
        mode=mode,  # Mode for downselecting the comparison points
        ndownsel=8,
        delay_method=delay_method)

    assert pos == approx(expect, abs=1e-2)


@pytest.fixture(params=["fit", "multi"])
def delay_method(request):
    return request.param


@pytest.mark.parametrize("delay", [-5.0, -2.5, -0.5, 0.0, 0.5, 2.5, 5.0])
def test_compute_delays(delay, delay_method):
    np.random.seed(2023)
    # Create a simple sinusoidal signal with noise
    fs = 500  # sample rate
    T = 1000.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    # duplicate and shift the signal
    y1 = np.roll(x, int(delay*fs))
    y2 = np.roll(x, int(2*delay * fs))
    y3 = np.roll(x, int(3*delay * fs))
    y4 = np.roll(x, int(4*delay * fs))

    df = pd.DataFrame(np.array([x,y1,y2,y3,y4]).T, index=pd.to_timedelta(t, 's'), columns=['x1','x2','x3','x4','x5'])
    ref = 'x1'

    delays, coh = field.compute_delays(df, ref, navgs=5, coh_limit=0.6, freq_limit=1, method=delay_method)

    assert (delays == approx([0, delay, 2*delay, 3*delay, 4*delay], abs=2e-3))


def test_compute_delays_nan(delay_method):
    delay = 5
    np.random.seed(2023)
    # Create a simple sinusoidal signal with noise
    fs = 500  # sample rate
    T = 1000.0    # seconds
    t = np.linspace(0, T, int(T*fs), endpoint=False)  # time variable

    x = 0.5 * np.sin(2 * np.pi * 2 * t) + np.random.random(len(t))

    # duplicate and shift the signal
    y1 = np.roll(x, int(delay*fs))
    y2 = np.roll(x, int(2*delay * fs))
    y3 = np.roll(x, int(3*delay * fs))
    y4 = np.nan * np.zeros_like(y3)

    df = pd.DataFrame(np.array([x,y1,y2,y3,y4]).T, index=pd.to_timedelta(t, 's'), columns=['x1','x2','x3','x4','x5'])
    ref = 'x1'

    delays, coh = field.compute_delays(df, ref, navgs=5, coh_limit=0.6, freq_limit=1, method=delay_method)

    assert (delays[0:-1] == approx([0, delay, 2*delay, 3*delay], abs=2e-3))
    assert np.isnan(delays[-1])


@pytest.fixture
def position_data():
    return pd.DataFrame({
        'E': [1, 2, 3, 4, 5],
        'N': [6, 7, 8, 9, 10]
    }, index=['A', 'B', 'C', 'D', 'E'])


@pytest.fixture
def remap_reverse():
    return [('A', 'E'), ('B', 'D'), ('C', 'C'), ('D', 'B'), ('E', 'A')]


@pytest.fixture
def remap_copy():
    return [('A', 'A'), ('B', 'B'), ('C', 'C'), ('D', 'D'), ('E', 'E')]


def test_assign_positions_types(position_data):
    remap_indices, data_out = field.assign_positions(position_data,
                                                     position_data)

    assert isinstance(remap_indices, list)
    assert all(isinstance(item, tuple) for item in remap_indices)
    assert isinstance(data_out, pd.DataFrame)

    assert len(remap_indices) == len(position_data)
    assert set(data_out.index) == set(position_data.index)


def test_assign_positions_duplicate(position_data):
    remap_indices, data_out = field.assign_positions(position_data,
                                                     position_data)

    expected_remap = list(zip(position_data.index,
                              position_data.index))
    assert remap_indices == expected_remap


def test_assign_positions_reverse(position_data):
    # Check whether we can reverse the order of the positions and pick that up
    predicted_pos = position_data.iloc[::-1]
    remap_indices, data_out = field.assign_positions(position_data,
                                                     predicted_pos)

    expected_remap = list(zip(position_data.index,
                              reversed(position_data.index)))
    assert remap_indices == expected_remap


def test_remap_data(position_data, remap_reverse):
    data_out = field.remap_positions(position_data, remap_reverse)

    assert isinstance(data_out, pd.DataFrame)

    for original, remapped in remap_reverse:
        assert data_out.loc[original].equals(position_data.loc[remapped])


def test_cascade_remap_nochange():
    remap1 = [('A', 'B'), ('B', 'C'), ('C', 'A'), ('D', 'D')]
    remap2 = [('A', 'A'), ('B', 'B'), ('C', 'C'), ('D', 'D')]

    result = field.cascade_remap(remap1, remap2)

    assert isinstance(result, list)
    assert all(isinstance(item, tuple) for item in result)

    expected_result = [('A', 'B'), ('B', 'C'), ('C', 'A'), ('D', 'D')]
    assert result == expected_result


def test_cascade_remap_changes():
    remap1 = [('A', 'B'), ('B', 'C'), ('C', 'A'), ('D', 'D')]
    remap2 = [('A', 'B'), ('B', 'C'), ('C', 'A'), ('D', 'D')]

    result = field.cascade_remap(remap1, remap2)

    assert isinstance(result, list)
    assert all(isinstance(item, tuple) for item in result)

    expected_result = [('A', 'C'), ('B', 'A'), ('C', 'B'), ('D', 'D')]
    assert result == expected_result



import pandas as pd
import os

from solarspatialtools.dataio import iotools

"""
Tool for processing NRCAN spatially resolved data as provided on the NRCAN 
data site. 

High Resolution Solar Datasets
   https://www.nrcan.gc.ca/energy/renewable-electricity/solar-photovoltaic/18409

Input
===================
The data should be provided as an individual csv file for each measurement 
sensor within the campaign. Those data need to be manually downloaded and 
stored in an individual directory for that particular day. The files must 
retain the original names with the dataset in order to format the dates, e.g. 
20140717_VAR01.csv. No other CSV files may be located within the directory. 
For example:

    Directory listing for c: \ data \ 20140717_Varennes \ 
    
        - 20140717_VAR01.csv
        - 20140717_VAR02.csv
        - ...

Because info about the sensor locations is not stored in the original data, 
a separate latlon CSV file is required. I had to create these manually. The 
format should be three column with Site, Latitude and Longitude columns. The
column names matter! For example:

    Contents of c: \ data \ Varennes_latlon.csv
     
        Site	Latitude	Longitude
        AFN01	44.190159	-78.096701
     
Output 
===================
The structure of the h5 files will be pandas DataFrame objects each with a key 
assigned. The keys will be as follows:

"data" - DataFrame has an index of the time, localized to UTC and contains the
irradiance in W/m2. Columns are labelled by station id.

"data_tilt" - DataFrame has an index of the time, localized to UTC and contains 
the tilted irradiance in W/m2. Columns are labelled by station id.

"latlon" - DataFrame will have an index of each stationid and columns "lat" and
"lon" for each station, measured in degrees East/North.

Parameters
===================
The parameters for this file are:
source_dir - a directory where all the CSV files for the given day are stored
latlon_file - the full path name of the latlon CSV file
target_file - the full path name of the output h5 file

See the bottom of the file for an example of usage



## Common format for H5 files used for Data Storage

I've tried to format the multisite time series measurements in a way that's 
conveinent for loading the files and working with the data. This came about 
from my initial work analyzing the HOPE Campaign, which used 100 individual 
point measurements of GHI scattered through a region near Jülich, Germany.

All data is collected into a single H5 file containing multiple fields. I use
```pandas``` and specifically ```pandas.read_hdf()``` for getting the data
into python. 

- ```latlon```: The latitude/longitude of the individual measurement sites
- ```utm```: The UTM coordinates of the individual measurement sites
- ```data```: Global Horizontal Irradiance
- ```data_tilt```: Global Tilted Irradiance (if available)

#### Location Data
Data about the location of each individual site is stored in the H5 file. Two
possible keys are used depending on the projection. Both are available when 
possible. The key ```latlon``` represents the site in a latitude coordinate
system. The key ```utm``` will contain the positions using UTM (or similar) 
projection that attempts to place the layout into a rectilinear coordinates. 
Upon use of ```pandas.read_hdf()``` the data will be brought into a DataFrame 
object.

- The index of the DataFrame is the site id. The HOPE datasets use an integer 
for the id, while NRCAN uses a string. 
- Columns are labelled ```lat``` and ```lon``` and contain the lat and lon in 
degrees for each of the distributed sensors (or ```E```, ```N``` in the case of 
```utm```).

#### Irradiance Data
Measurements consist of the individual sensor time series with a shared time 
index. Upon use of ```pandas.read_hdf()``` the data will be brought into a 
DataFrame object. Each individual sensor has its own column. 

- Index of the DataFrame is the timestamp referenced to a timezone
- Columns contain the time series for each individual sensor, and are keyed by
the site id (HOPE - integer, NRCAN - string).

"""


def process_files(source_dir, latlon_file, target_file):
    """
    Perform the file conversion on NRCAN data into an h5 file with consistent
    time step.

    Parameters
    ----------
    source_dir : string
        a directory where all the CSV files for the given day are stored
    latlon_file : string
        the full path name of the latlon CSV file
    target_file: string
        the full path name of the output h5 file

    Output
    -------
        an h5 file saved to disk with keys 'data','data_tilt' and 'latlon'
    """

    # All the files in the source directory
    files = iotools.getfiles(source_dir, ".csv")

    # process the lat lon data
    latlon = pd.read_csv(latlon_file)
    latlon = latlon.rename(columns={"Site": "id",
                                    "Latitude": "lat",
                                    "Longitude": "lon"})
    latlon = latlon.set_index('id')

    # pass through all the files
    for i, file in enumerate(files):
        if "latlon" in file and ".csv" in file:
            pass
        else:
            # File for a single sensor
            data = pd.read_csv(os.path.join(source_dir, file))

            # Generate the time stamps
            file_time = pd.to_datetime(data.iloc[:, 0])\
                        + pd.to_timedelta(data.iloc[:, 1])
            #  Uses GMT+5, rather than Eastern time with DST shift.
            file_time = pd.DatetimeIndex(file_time).tz_localize('Etc/GMT+5')

            # We are currently converting this sensor's data into a column
            col = os.path.splitext(file)[0].split('_')[1]

            # Extract the GHI
            ghi = pd.DataFrame(data["G1 (W/m2)"]).rename(columns=
                                                         {'G1 (W/m2)': col})
            ghi.index = file_time

            # Fill in blank data for even sampling
            ghi = ghi.resample("10ms").ffill()

            # Build the GHI dataframe
            if 'data_ghi' not in vars():
                data_ghi = pd.DataFrame(ghi, index=ghi.index, columns=[col])
            else:
                data_ghi = data_ghi.assign(tmp=ghi).rename(columns=
                                                           {'tmp': col})

            # Extract the Global Tilted Irradiance
            gti = pd.DataFrame(data["G2 (W/m2)"]).rename(columns=
                                                         {'G2 (W/m2)': col})
            gti.index = file_time

            # Fill in blank data
            gti = gti.resample("10ms").ffill()

            # Build the GTI dataframe
            if 'data_gti' not in vars():
                data_gti = pd.DataFrame(gti, index=gti.index, columns=[col])
            else:
                data_gti = data_gti.assign(tmp=gti).rename(columns=
                                                           {'tmp': col})

    # All the dataframes are built, save them to the h5 file
    latlon.to_hdf(target_file, key='latlon', mode='a')
    data_ghi.to_hdf(target_file, key='data', mode='a', append=True)
    data_gti.to_hdf(target_file, key='data_tilt', mode='a', append=True)


if __name__ == "__main__":

    # Build the paths for the individual datafiles
    rootdir = 'c:/path/to/data'

    filedir = '20151008_Alderville'
    site = "Alderville"
    target_fn = 'nrcan_ald_v.h5'

    # filedir = '20150812_Alderville'
    # site = "Alderville"
    # target_fn = 'nrcan_ald_hv.h5'

    # filedir = '20140717_Varennes'
    # site = "Varennes"
    # target_fn = 'nrcan_var_v.h5'

    # filedir = '20150226_Varennes'
    # site = "Varennes"
    # target_fn = 'nrcan_var_hv.h5'

    target_rootdir = 'c:/path/to/output/'

    source_dir = os.path.join(rootdir, filedir)
    latlon_file = os.path.join(rootdir, f"latlon_{site}.csv")
    target_file = os.path.join(target_rootdir, target_fn)

    process_files(source_dir, latlon_file, target_file)



import os.path
import argparse
import pandas as pd
import numpy as np
import netCDF4
from solarspatialtools.dataio.iotools import ensurepath, getfiles
from solarspatialtools.dataio.iotools import download, wget_fromurl

"""
Tool for downloading all of the HOPE and HOPE-Melpitz data from the University
of Hamburg data site. Replaces use of their provided wget files for a Windows
environment. 
Old Site: 
   https://icdc.cen.uni-hamburg.de/all-samd-data.html
   
   Files can be navigated to via:
     Short Term Observations > HOPE > Instrument Groups > ...
         Radiation & Imager > rsds    ---> wget_hope.sh

     Short Term Observations > HOPE Melpitz > Instrument Groups > ...
         Radiation & Imager > rsds   ---> wget_melpitz.sh
New Site:
   Data > Atmosphere > SAMD Data Sets - Short Term Observations > HOPE 
   Field for irradiance is "rsds"
   https://www.cen.uni-hamburg.de/en/icdc/data/atmosphere/samd-st-datasets.html

If generated, the structure of the h5 files will be pandas DataFrame objects 
each with a key assigned. The three keys will be as follows:

"data" - DataFrame has an index of the time, localized to UTC and contains the
irradiance in W/m2. Columns are labelled by station id.

"flag" -  DataFrame has an index of the time, localized to UTC and contains the
data flag. Columns are labelled by station id. According to the raw field info 
inside the raw files, the flag meanings are:
    1 - good_data
    2 - okey_but_sometimes_dubious_data
    3 - bad_data_ignore_completely
    4 - no_observation

"latlon" - DataFrame will have an index of each stationid and columns "lat" and
"lon" for each station, measured in degrees East/North.


example usage:
    python hope_campaign.py HOPE d:\data\hope -m -h5 -q interp
    
    
    
## Common format for H5 files used for Data Storage

I've tried to format the multisite time series measurements in a way that's 
conveinent for loading the files and working with the data. This came about 
from my initial work analyzing the HOPE Campaign, which used 100 individual 
point measurements of GHI scattered through a region near Jülich, Germany.

All data is collected into a single H5 file containing multiple fields. I use
```pandas``` and specifically ```pandas.read_hdf()``` for getting the data
into python. 

- ```latlon```: The latitude/longitude of the individual measurement sites
- ```utm```: The UTM coordinates of the individual measurement sites
- ```data```: Global Horizontal Irradiance
- ```data_tilt```: Global Tilted Irradiance (if available)

#### Location Data
Data about the location of each individual site is stored in the H5 file. Two
possible keys are used depending on the projection. Both are available when 
possible. The key ```latlon``` represents the site in a latitude coordinate
system. The key ```utm``` will contain the positions using UTM (or similar) 
projection that attempts to place the layout into a rectilinear coordinates. 
Upon use of ```pandas.read_hdf()``` the data will be brought into a DataFrame 
object.

- The index of the DataFrame is the site id. The HOPE datasets use an integer 
for the id, while NRCAN uses a string. 
- Columns are labelled ```lat``` and ```lon``` and contain the lat and lon in 
degrees for each of the distributed sensors (or ```E```, ```N``` in the case of 
```utm```).

#### Irradiance Data
Measurements consist of the individual sensor time series with a shared time 
index. Upon use of ```pandas.read_hdf()``` the data will be brought into a 
DataFrame object. Each individual sensor has its own column. 

- Index of the DataFrame is the timestamp referenced to a timezone
- Columns contain the time series for each individual sensor, and are keyed by
the site id (HOPE - integer, NRCAN - string).

"""

# Magic strings pointing to the files for each of the HOPE data sets
URLS = {"HOPE":
        {"wget": "https://icdc.cen.uni-hamburg.de/thredds/fileServer/"
                 "ftpthredds/samd/"
                 "data/pyrnet/00/rsds/l1/hope/trop/2013/wget.sh",
         "meta": "https://icdc.cen.uni-hamburg.de/thredds/fileServer/"
                 "ftpthredds/samd/"
                 "metadata/icdc/hope_trop_pyrnet00_l1_rsds_v00.xml"},
        "HOPE-MELPITZ":
        {"wget": "https://icdc.cen.uni-hamburg.de/thredds/fileServer/"
                 "ftpthredds/samd/"
                 "data/pyrnet/00/rsds/l1/hopm/trop/2013/wget.sh",
         "meta": "https://icdc.cen.uni-hamburg.de/thredds/fileServer/"
                 "ftpthredds/samd/"
                 "metadata/icdc/hopm_trop_pyrnet00_l1_rsds_v00.xml"},
        "LINDENBERG":
        {"wget": "https://icdc.cen.uni-hamburg.de/thredds/fileServer/"
                 "ftpthredds/samd/"
                 "data/pyrnet/00/rsds/l1/ioprao/trop/2014/wget.sh",
         "meta": "https://icdc.cen.uni-hamburg.de/thredds/fileServer/"
                 "ftpthredds/samd/"
                 "metadata/icdc/ioprao_trop_pyrnet00_l1_rsds_v00.xml"}}


def flag_qc(data, flag, level=2):
    """

    Parameters
    ----------
    data : pandas DataFrame
        A dataframe containing the data to interpolate

    flag : pandas DataFrame
        A dataframe containing the flags for the data

    level : int
        The lowest unacceptable flag level (flag >= level is set to nan)

    Returns
    -------
    datanew : pandas DataFrame
        A modified version of the data
    """
    datanew = data.copy()
    datanew[flag >= level] = np.nan
    return datanew


def null_qc(data):
    """
    Replace data with nan based on the dataset hidden values (-999)

    Parameters
    ----------
    data : pandas DataFrame
        A dataframe containing the data to limit

    Returns
    -------
    datanew : pandas DataFrame
        A modified version of the data
    """

    datanew = data.copy()
    datanew[datanew < -100] = np.nan
    return datanew


def interp_nan(data, limit=5):
    """
    Interpolate over nan values. Duration is limited by the limit parameter.
    Periods of nan values that span longer than the limit will be left as nan.

    Parameters
    ----------
    data : pandas DataFrame
        A dataframe containing the data to interpolate

    limit : int, default 5
        The limit to the number of timesteps to interpolate through

    Returns
    -------
    datanew : pandas DataFrame
        A modified version of the data
    """

    datanew = data.copy()

    for col in datanew.columns:
        datanew_i = datanew[col]
        # interpolate limiting things to 5s or shorter nan windows
        datanew_i = datanew_i.interpolate(limit=limit, limit_area='inside')

        # Find the cases where things were nan for more than limit, and add
        # some padding to ensure we don't interp there
        blanks = np.isnan(datanew_i).rolling(f'{limit}s').mean() > \
                 (limit-1)/limit
        blanks_front = blanks.iloc[:-2*limit]
        blanks_back = blanks.iloc[2*limit:]
        blanks_union = blanks_front.values | blanks_back.values
        blanks = pd.Series(np.pad(blanks_union, limit), index=blanks.index)
        datanew_i[blanks] = np.nan

        datanew[col] = datanew_i

    return datanew


def parse_file(filename):
    """
    Convert a datafile in the HOPE format to pandas DataFrame objects.

    data DataFrame has an index of the time, localized to UTC and contains the
    irradiance in W/m2. Columns are labelled by station id.

    flag DataFrame has an index of the time, localized to UTC and contains the
    data flag. Columns are labelled by station id. According to the data
    description PDF file, the flag meanings are:
        1 - good_data
        2 - okey_but_sometimes_dubious_data
        3 - bad_data_ignore_completely
        4 - no_observation

    The stationdat DataFrame will have an index of each stationid and columns
    "lat" and "lon" for each station, measured in degrees East/North.

    Parameters
    ----------
    filename : string
        the full path filename to the datafile

    Returns
    -------
    (data, flag, stationdat) : (pd.DataFrame, pd.DataFrame, pd.DataFrame)
        data is a dataframe of the actual time series indexed by timestamp and
            with columns of each station id
        flag is a dataframe of the qc flags indexed by timestamp and with
            columns of each station id
        stationid is a dataframe with the location of each of the individual
            stations indexed by station id and with columns of "lat" and "lon"
            in degrees.
    """

    # Read file
    data = netCDF4.Dataset(filename, mode='r')
    data.set_auto_mask(False)

    # Station Identifiers
    ids = data['station_id'][:]

    # The station info Frame
    lat = data['lat'][:]
    lon = data['lon'][:]

    stationdat = pd.DataFrame(np.array([lat, lon]).transpose(),
                              index=ids.transpose(),
                              columns=['lat', 'lon'])

    # Get all times, convert to pandas time and localize to UTC
    time = data['time']
    timeindex = pd.to_datetime(time[:], unit="s")
    timeindex = timeindex.tz_localize('UTC')

    # Get Irradiance
    irrad = data['rsds'][:]
    irrad_flag = data['rsds_flag'][:]

    # Build Data and Flag Frames
    data = pd.DataFrame(irrad, columns=ids, index=timeindex)
    flag = pd.DataFrame(irrad_flag, columns=ids, index=timeindex)

    return data, flag, stationdat


def dataset_to_h5(directory, filename, overwrite=False, verbose=True):
    """
    Convert a dataset stored in a directory to h5 file. DataFrame objects each
    have a key assigned. The three keys will be as follows:

    "data" - DataFrame has an index of the time, localized to UTC and contains
    the irradiance in W/m2. Columns are labelled by station id.

    "flag" -  DataFrame has an index of the time, localized to UTC and contains
    the data flag. Columns are labelled by station id. According to the raw
    field info inside the raw files, the flag meanings are:
    1 - good_data
    2 - okey_but_sometimes_dubious_data
    3 - bad_data_ignore_completely
    4 - no_observation

    "latlon" - DataFrame will have an index of each stationid and columns "lat"
    and "lon" for each station, measured in degrees East/North.

    Parameters
    ----------
    directory : string
        directory holding the downloaded dataset files

    filename : string
        Full context filename of the target h5 file.

    overwrite : bool
        should the file be overwritten if it exists?

    verbose : bool
        print status updates?

    Returns
    -------
    void
    """

    files = getfiles(directory, ext=".nc", sort=True)

    if os.path.isfile(filename):
        if overwrite:
            os.remove(filename)
        else:
            raise ValueError("Target file exists. Please specify a new file.")

    for i, fn in enumerate(files):
        if verbose:
            print(fn)

        data, flag, latlon = parse_file(os.path.join(directory, fn))

        if i == 0:
            latlon.to_hdf(filename, key='latlon', mode="a")

        data.to_hdf(filename, key='data', mode='a', append=True)
        flag.to_hdf(filename, key='flag', mode='a', append=True)


def _parse_args():
    """
    parse the arguments specified in the program call.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("dataset", help="Name of the dataset to get")
    parser.add_argument("target", help="Target directory to save data")
    parser.add_argument("-m", "--meta", help="Download metadata?",
                        action="store_true")
    parser.add_argument("-h5", "--makeh5", help="create h5 summary file",
                        action="store_true")
    parser.add_argument("-s", "--silent", help="Hide printouts?",
                        action="store_true")
    parser.add_argument("-q", "--qc", help="Quality control.")
    args = parser.parse_args()
    if args.dataset.upper() not in URLS.keys():
        raise ValueError("Dataset must be one of: " + str(URLS.keys()))
    return args


def _main():
    args = _parse_args()
    ensurepath(args.target)
    if args.meta:
        if not args.silent:
            print("Metadata")
        download(URLS[args.dataset.upper()]['meta'],
                 os.path.join(args.target, args.dataset + "_metadata.xml"))
    wget_fromurl(URLS[args.dataset.upper()]['wget'],
                 args.target, not args.silent)

    if args.makeh5:
        try:
            dataset_to_h5(args.target,
                          os.path.join(args.target,
                                       args.dataset.lower() + ".h5"),
                          overwrite=False,
                          verbose=(not args.silent))
        except ValueError:
            userinput = input("File exists. Overwrite (Y/N)?")
            if userinput.lower() == "y":
                dataset_to_h5(args.target,
                              os.path.join(args.target,
                                           args.dataset.lower() + ".h5"),
                              overwrite=True,
                              verbose=(not args.silent))
            else:
                print("Continuing")

    if args.qc is not None:
        print('qc')
        fn_src = os.path.join(args.target, args.dataset.lower() + ".h5")
        fn_tgt = os.path.join(args.target,
                     args.dataset.lower() + "-qc{}.h5".format(args.qc.lower()))
        data = pd.read_hdf(fn_src, mode="r", key="data")
        flag = pd.read_hdf(fn_src, mode="r", key="flag")
        latlon = pd.read_hdf(fn_src, mode="r", key="latlon")

        if args.qc == "interp":
            data = null_qc(data)
            data = interp_nan(data)
        elif args.qc == 'null':
            data = null_qc(data)
        elif args.qc == 'flag':
            data = flag_qc(data, flag, 2)
        else:
            raise ValueError('QC must be one of None, null, interp, flag.')
        if os.path.isfile(fn_tgt):
            userinput = input("File exists. Overwrite (Y/N)?")
            if userinput.lower() == "y":
                os.remove(fn_tgt)
            else:
                return
        data.to_hdf(fn_tgt, key='data', mode='a', append=False)
        flag.to_hdf(fn_tgt, key='flag', mode='a', append=False)
        latlon.to_hdf(fn_tgt, key='latlon', mode='a', append=False)


if __name__ == "__main__":
    _main()
